# Copyright (c) 2010-2016 Emmanuel Blot <emmanuel.blot@free.fr>
# Copyright (c) 2008-2015, Neotion
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

"""Serial modules compliant with pyserial APIs
"""

try:
    from serial.serialutil import SerialException
except ImportError as exc:
    raise ImportError("Python serial module not installed") from exc
try:
    from serial import VERSION, serial_for_url as serial4url
    version = tuple([int(x) for x in VERSION.split('.')])
    if version < (3, 0):
        raise ValueError
except (ValueError, IndexError, ImportError) as exc:
    raise ImportError("pyserial 3.0+ is required") from exc
try:
    from serial import protocol_handler_packages
    protocol_handler_packages.append('pyftdi.serialext')
except ImportError as exc:
    raise SerialException('Cannot register pyftdi extensions') from exc

serial_for_url = serial4url


def touch():
    """Do nothing, only for static checkers than do not like module import
       with no module references
    """



"""PyUSB virtual FTDI device."""

# Copyright (c) 2020, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

from collections import deque
from logging import getLogger
from struct import unpack as sunpack
from typing import Union
from pyftdi.tracer import FtdiMpsseEngine, FtdiMpsseTracer


class VirtMpsseTracer(FtdiMpsseTracer):
    """Reuse MPSSE tracer as a MPSSE command decoder engine.
    """

    def __init__(self, port: 'VirtFtdiPort', version: int):
        super().__init__(version)
        self.log = getLogger('pyftdi.virt.mpsse.{port.iface}')
        self._port = port

    def _get_engine(self, iface: int):
        iface -= 1
        try:
            self._engines[iface]
        except IndexError as exc:
            raise ValueError('No MPSSE engine available on interface %d' %
                             iface) from exc
        if not self._engines[iface]:
            self._engines[iface] = VirtMpsseEngine(self, self._port)
        return self._engines[iface]


class VirtMpsseEngine(FtdiMpsseEngine):
    """Virtual implementation of a MPSSE.

       Far from being complete for now :-)
    """

    def __init__(self, tracer: VirtMpsseTracer, port: 'VirtFtdiPort'):
        super().__init__(port.iface)
        self.log = getLogger(f'pyftdi.virt.mpsse.{port.iface}')
        self._tracer = tracer
        self._port = port
        self._width = port.width
        self._mask = (1 << self._width) - 1
        self._reply_q = deque()

    def send(self, buf: Union[bytes, bytearray]) -> None:
        super().send(buf)
        # cannot post the response before the send() method has completed
        # see FtdiMpsseEngine.send() for execution steps: expected reply size
        # is only known (stored) once the command execution has completed
        self.reply()

    def reply(self) -> None:
        """Post the reply to a command back into the virtual FTDI FIFO."""
        while self._reply_q:
            self._port.write_from_mpsse(self, self._reply_q.popleft())

    def _cmd_get_bits_low(self):
        super()._cmd_get_bits_low()
        byte = self._port.gpio & 0xff
        buf = bytes([byte])
        self._reply_q.append(buf)
        return True

    def _cmd_get_bits_high(self):
        super()._cmd_get_bits_high()
        byte = (self._port.gpio >> 8) & 0xff
        buf = bytes([byte])
        self._reply_q.append(buf)
        return True

    def _cmd_set_bits_low(self):
        buf = self._trace_tx[1:3]
        if not super()._cmd_set_bits_low():
            return False
        port = self._port
        byte, direction = sunpack('BB', buf)
        gpi = port.gpio & ~direction & self._mask
        gpo = byte & direction & self._mask
        msb = port.gpio & ~0xFF
        gpio = gpi | gpo | msb
        port.update_gpio(self, False, direction, gpio)
        self.log.debug('. bbwl %04x: %s', port.gpio, f'{port.gpio:016b}')
        return True

    def _cmd_set_bits_high(self):
        buf = self._trace_tx[1:3]
        if not super()._cmd_set_bits_high():
            return False
        port = self._port
        byte, direction = sunpack('BB', buf)
        byte <<= 8
        direction <<= 8
        gpi = port.gpio & ~direction & self._mask
        gpo = byte & direction & self._mask
        lsb = port.gpio & 0xFF
        gpio = gpi | gpo | lsb
        port.update_gpio(self, False, direction, gpio)
        self.log.debug('. bbwh %04x: %s', port.gpio, f'{port.gpio:016b}')
        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2020-2021, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.

#pylint: disable-msg=empty-docstring
#pylint: disable-msg=missing-docstring
#pylint: disable-msg=no-self-use
#pylint: disable-msg=invalid-name
#pylint: disable-msg=global-statement
#pylint: disable-msg=too-many-locals


import logging
from collections import defaultdict
from contextlib import redirect_stdout
from doctest import testmod
from io import StringIO
from os import environ
from string import ascii_letters
from sys import modules, stdout, version_info
from unittest import TestCase, TestSuite, makeSuite, main as ut_main
from urllib.parse import urlsplit
from pyftdi import FtdiLogger
from pyftdi.eeprom import FtdiEeprom
from pyftdi.ftdi import Ftdi, FtdiMpsseError
from pyftdi.gpio import GpioController
from pyftdi.misc import to_bool
from pyftdi.serialext import serial_for_url
from pyftdi.usbtools import UsbTools

# MockLoader is assigned in ut_main
MockLoader = None


class FtdiTestCase(TestCase):
    """Common features for all tests.
    """

    @classmethod
    def setUpClass(cls):
        cls.loader = MockLoader()
        cls.debug = to_bool(environ.get('FTDI_DEBUG', 'off'), permissive=False)

    @classmethod
    def tearDownClass(cls):
        if cls.loader:
            cls.loader.unload()

    def setUp(self):
        if self.debug:
            print('.'.join(self.id().split('.')[-2:]))


class MockUsbToolsTestCase(FtdiTestCase):
    """Test UsbTools APIs.
    """

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        with open('pyftdi/tests/resources/ftmany.yaml', 'rb') as yfp:
            cls.loader.load(yfp)

    def test_enumerate(self):
        """Enumerate FTDI devices."""
        ftdis = [(0x403, pid)
                 for pid in (0x6001, 0x6010, 0x6011, 0x6014, 0x6015)]
        count = len(UsbTools.find_all(ftdis))
        self.assertEqual(count, 6)

    def test_device(self):
        """Access and release FTDI device."""
        ftdis = [(0x403, 0x6001)]
        ft232rs = UsbTools.find_all(ftdis)
        self.assertEqual(len(ft232rs), 1)
        devdesc, ifcount = ft232rs[0]
        self.assertEqual(ifcount, 1)
        dev = UsbTools.get_device(devdesc)
        self.assertIsNotNone(dev)
        UsbTools.release_device(dev)

    def test_string(self):
        """Retrieve a string from its identifier."""
        ftdis = [(0x403, 0x6010)]
        ft2232h = UsbTools.find_all(ftdis)[0]
        devdesc, _ = ft2232h
        dev = UsbTools.get_device(devdesc)
        serialn = UsbTools.get_string(dev, dev.iSerialNumber)
        self.assertEqual(serialn, 'FT2DEF')

    def test_list_devices(self):
        """List FTDI devices."""
        vid = 0x403
        vids = {'ftdi': vid}
        pids = {
            vid: {
                '230x': 0x6015,
                '232r': 0x6001,
                '232h': 0x6014,
                '2232h': 0x6010,
                '4232h': 0x6011,
            }
        }
        devs = UsbTools.list_devices('ftdi:///?', vids, pids, vid)
        self.assertEqual(len(devs), 6)
        ifmap = {
            0x6001: 1,
            0x6010: 2,
            0x6011: 4,
            0x6014: 1,
            0x6015: 1
        }
        for dev, desc in devs:
            strings = UsbTools.build_dev_strings('ftdi', vids, pids,
                                                 [(dev, desc)])
            self.assertEqual(len(strings), ifmap[dev.pid])
            for url, _ in strings:
                parts, _ = UsbTools.parse_url(url, 'ftdi', vids, pids, vid)
                self.assertEqual(parts.vid, dev.vid)
                self.assertEqual(parts.pid, dev.pid)
                self.assertEqual(parts.bus, dev.bus)
                self.assertEqual(parts.address, dev.address)
                self.assertEqual(parts.sn, dev.sn)
        devs = UsbTools.list_devices('ftdi://:232h/?', vids, pids, vid)
        self.assertEqual(len(devs), 2)
        devs = UsbTools.list_devices('ftdi://:2232h/?', vids, pids, vid)
        self.assertEqual(len(devs), 1)


class MockFtdiDiscoveryTestCase(FtdiTestCase):
    """Test FTDI device discovery APIs.
       These APIs are FTDI wrappers for UsbTools APIs.
    """

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        with open('pyftdi/tests/resources/ftmany.yaml', 'rb') as yfp:
            cls.loader.load(yfp)

    def test_list_devices(self):
        """List FTDI devices."""
        devs = Ftdi.list_devices('ftdi:///?')
        self.assertEqual(len(devs), 6)
        devs = Ftdi.list_devices('ftdi://:232h/?')
        self.assertEqual(len(devs), 2)
        devs = Ftdi.list_devices('ftdi://:2232h/?')
        self.assertEqual(len(devs), 1)
        devs = Ftdi.list_devices('ftdi://:4232h/?')
        self.assertEqual(len(devs), 1)
        out = StringIO()
        Ftdi.show_devices('ftdi:///?', out)
        lines = [l.strip() for l in out.getvalue().split('\n')]
        lines.pop(0)  # "Available interfaces"
        while lines and not lines[-1]:
            lines.pop()
        self.assertEqual(len(lines), 10)
        portmap = defaultdict(int)
        reference = {'232': 1, '2232': 2, '4232': 4, '232h': 2, 'ft-x': 1}
        for line in lines:
            url = line.split(' ')[0].strip()
            parts = urlsplit(url)
            self.assertEqual(parts.scheme, 'ftdi')
            self.assertRegex(parts.path, r'^/[1-4]$')
            product = parts.netloc.split(':')[1]
            portmap[product] += 1
        self.assertEqual(portmap, reference)


class MockSimpleDeviceTestCase(FtdiTestCase):
    """Test FTDI APIs with a single-port FTDI device (FT232H)
    """

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        with open('pyftdi/tests/resources/ft232h.yaml', 'rb') as yfp:
            cls.loader.load(yfp)

    def test_enumerate(self):
        """Check simple enumeration of a single FTDI device."""
        ftdi = Ftdi()
        temp_stdout = StringIO()
        with redirect_stdout(temp_stdout):
            self.assertRaises(SystemExit, ftdi.open_from_url, 'ftdi:///?')
        lines = [l.strip() for l in temp_stdout.getvalue().split('\n')]
        lines.pop(0)  # "Available interfaces"
        while lines and not lines[-1]:
            lines.pop()
        self.assertEqual(len(lines), 1)
        self.assertTrue(lines[0].startswith('ftdi://'))
        # skip description, i.e. consider URL only
        self.assertTrue(lines[0].split(' ')[0].endswith('/1'))


class MockDualDeviceTestCase(FtdiTestCase):
    """Test FTDI APIs with two similar single-port FTDI devices (FT232H)
    """

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        with open('pyftdi/tests/resources/ft232h_x2.yaml', 'rb') as yfp:
            cls.loader.load(yfp)

    def test_enumerate(self):
        """Check simple enumeration of a 2-port FTDI device."""
        ftdi = Ftdi()
        temp_stdout = StringIO()
        with redirect_stdout(temp_stdout):
            self.assertRaises(SystemExit, ftdi.open_from_url, 'ftdi:///?')
        lines = [l.strip() for l in temp_stdout.getvalue().split('\n')]
        lines.pop(0)  # "Available interfaces"
        while lines and not lines[-1]:
            lines.pop()
        self.assertEqual(len(lines), 2)
        for line in lines:
            self.assertTrue(line.startswith('ftdi://'))
            # skip description, i.e. consider URL only
            self.assertTrue(line.split(' ')[0].endswith('/1'))


class MockTwoPortDeviceTestCase(FtdiTestCase):
    """Test FTDI APIs with a dual-port FTDI device (FT2232H)
    """

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        with open('pyftdi/tests/resources/ft2232h.yaml', 'rb') as yfp:
            cls.loader.load(yfp)

    def test_enumerate(self):
        """Check simple enumeration of a 4-port FTDI device."""
        ftdi = Ftdi()
        temp_stdout = StringIO()
        with redirect_stdout(temp_stdout):
            self.assertRaises(SystemExit, ftdi.open_from_url, 'ftdi:///?')
        lines = [l.strip() for l in temp_stdout.getvalue().split('\n')]
        lines.pop(0)  # "Available interfaces"
        while lines and not lines[-1]:
            lines.pop()
        self.assertEqual(len(lines), 2)
        for pos, line in enumerate(lines, start=1):
            self.assertTrue(line.startswith('ftdi://'))
            # skip description, i.e. consider URL only
            self.assertTrue(line.split(' ')[0].endswith(f'/{pos}'))


class MockFourPortDeviceTestCase(FtdiTestCase):
    """Test FTDI APIs with a quad-port FTDI device (FT4232H)
    """

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        with open('pyftdi/tests/resources/ft4232h.yaml', 'rb') as yfp:
            cls.loader.load(yfp)

    def test_enumerate(self):
        """Check simple enumeration of two similar FTDI device."""
        ftdi = Ftdi()
        temp_stdout = StringIO()
        with redirect_stdout(temp_stdout):
            self.assertRaises(SystemExit, ftdi.open_from_url, 'ftdi:///?')
        lines = [l.strip() for l in temp_stdout.getvalue().split('\n')]
        lines.pop(0)  # "Available interfaces"
        while lines and not lines[-1]:
            lines.pop()
        self.assertEqual(len(lines), 4)
        for pos, line in enumerate(lines, start=1):
            self.assertTrue(line.startswith('ftdi://'))
            # skip description, i.e. consider URL only
            self.assertTrue(line.split(' ')[0].endswith(f'/{pos}'))


class MockManyDevicesTestCase(FtdiTestCase):
    """Test FTDI APIs with several, mixed type FTDI devices
    """

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        with open('pyftdi/tests/resources/ftmany.yaml', 'rb') as yfp:
            cls.loader.load(yfp)

    def test_enumerate(self):
        """Check simple enumeration of two similar FTDI device."""
        ftdi = Ftdi()
        temp_stdout = StringIO()
        with redirect_stdout(temp_stdout):
            self.assertRaises(SystemExit, ftdi.open_from_url, 'ftdi:///?')
        lines = [l.strip() for l in temp_stdout.getvalue().split('\n')]
        lines.pop(0)  # "Available interfaces"
        while lines and not lines[-1]:
            lines.pop()
        self.assertEqual(len(lines), 10)
        for line in lines:
            self.assertTrue(line.startswith('ftdi://'))
            # skip description, i.e. consider URL only
            url = line.split(' ')[0]
            urlparts = urlsplit(url)
            self.assertEqual(urlparts.scheme, 'ftdi')
            parts = urlparts.netloc.split(':')
            if parts[1] == '4232':
                # def file contains no serial number, so expect bus:addr syntax
                self.assertEqual(len(parts), 4)
                self.assertRegex(parts[2], r'^\d$')
                self.assertRegex(parts[3], r'^\d$')
            else:
                # other devices are assigned a serial number
                self.assertEqual(len(parts), 3)
                self.assertTrue(parts[2].startswith('FT'))
            self.assertRegex(urlparts.path, r'^/\d$')


class MockSimpleDirectTestCase(FtdiTestCase):
    """Test FTDI open/close APIs with a basic featured FTDI device (FT230H)
    """

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        with open('pyftdi/tests/resources/ft230x.yaml', 'rb') as yfp:
            cls.loader.load(yfp)

    def test_open_close(self):
        """Check simple open/close sequence."""
        ftdi = Ftdi()
        ftdi.open_from_url('ftdi:///1')
        self.assertEqual(ftdi.usb_path, (1, 1, 0))
        ftdi.close()

    def test_open_bitbang(self):
        """Check simple open/close BitBang sequence."""
        ftdi = Ftdi()
        ftdi.open_bitbang_from_url('ftdi:///1')
        ftdi.close()

    def test_open_mpsse(self):
        """Check simple MPSSE access."""
        ftdi = Ftdi()
        # FT230X is a pure UART bridge, MPSSE should not be available
        self.assertRaises(FtdiMpsseError,
                          ftdi.open_mpsse_from_url, 'ftdi:///1')


class MockSimpleMpsseTestCase(FtdiTestCase):
    """Test FTDI open/close APIs with a MPSSE featured FTDI device (FT232H)
    """

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        with open('pyftdi/tests/resources/ft232h.yaml', 'rb') as yfp:
            cls.loader.load(yfp)

    def test_open_close(self):
        """Check simple open/close sequence."""
        ftdi = Ftdi()
        ftdi.open_from_url('ftdi:///1')
        self.assertEqual(ftdi.usb_path, (4, 5, 0))
        ftdi.close()

    def test_open_bitbang(self):
        """Check simple open/close BitBang sequence."""
        ftdi = Ftdi()
        ftdi.open_bitbang_from_url('ftdi:///1')
        ftdi.close()

    def test_open_mpsse(self):
        """Check simple MPSSE access."""
        ftdi = Ftdi()
        ftdi.open_mpsse_from_url('ftdi:///1')
        ftdi.close()


class MockSimpleGpioTestCase(FtdiTestCase):
    """Test FTDI GPIO APIs
    """

    def tearDown(self):
        self.loader.unload()

    def _test_gpio(self):
        """Check simple GPIO write and read sequence."""
        with open('pyftdi/tests/resources/ft232h.yaml', 'rb') as yfp:
            self.loader.load(yfp)
        gpio = GpioController()
        # access to the virtual GPIO port
        out_pins = 0xAA
        gpio.configure('ftdi://:232h/1', direction=out_pins)
        bus, address, iface = gpio.ftdi.usb_path
        self.assertEqual((bus, address, iface), (4, 5, 0))
        vftdi = self.loader.get_virtual_ftdi(bus, address)
        vport = vftdi.get_port(1)
        gpio.write_port(0xF3)
        self.assertEqual(vport.gpio, 0xAA & 0xF3)
        vport.gpio = 0x0c
        vio = gpio.read_port()
        self.assertEqual(vio, (0xAA & 0xF3) | (~0xAA & 0x0c))
        gpio.close()

    def test_baudrate(self):
        """Check simple GPIO write and read sequence."""
        # load custom CBUS config, with:
            # CBUS0: GPIO (gpio)
            # CBUS1: GPIO (gpio)
            # CBUS0: DRIVE1 (forced to high level)
            # CBUS0: TXLED  (eq. to highz for tests)
        with open('pyftdi/tests/resources/ft230x_io.yaml', 'rb') as yfp:
            self.loader.load(yfp)
        gpio = GpioController()
        # access to the virtual GPIO port
        out_pins = 0xAA
        gpio.configure('ftdi://:230x/1', direction=out_pins)
        vftdi = self.loader.get_virtual_ftdi(1, 1)
        vftdi.get_port(1)
        baudrate = 1000000
        gpio.set_frequency(baudrate)
        gpio.close()


class MockSimpleUartTestCase(FtdiTestCase):
    """Test FTDI UART APIs
    """

    def tearDown(self):
        self.loader.unload()

    def test_uart(self):
        """Check simple TX/RX sequence."""
        with open('pyftdi/tests/resources/ft232h.yaml', 'rb') as yfp:
            self.loader.load(yfp)
        port = serial_for_url('ftdi:///1')
        bus, address, _ = port.usb_path
        vftdi = self.loader.get_virtual_ftdi(bus, address)
        vport = vftdi.get_port(1)
        msg = ascii_letters
        port.write(msg.encode())
        txd = vport[vport.UART_PINS.TXD]
        buf = txd.read(len(ascii_letters)+10).decode()
        self.assertEqual(msg, buf)
        msg = ''.join(reversed(msg))
        rxd = vport[vport.UART_PINS.TXD]
        rxd.write(msg.encode())
        buf = port.read(len(ascii_letters)).decode()
        self.assertEqual(msg, buf)
        port.close()

    def test_uart_loopback(self):
        """Check TXD/RXD loopback."""
        with open('pyftdi/tests/resources/ft232h.yaml', 'rb') as yfp:
            self.loader.load(yfp)
        port = serial_for_url('ftdi:///1')
        bus, address, _ = port.usb_path
        vftdi = self.loader.get_virtual_ftdi(bus, address)
        vport = vftdi.get_port(1)
        txd = vport[vport.UART_PINS.TXD]
        rxd = vport[vport.UART_PINS.RXD]
        txd.connect_to(rxd)
        msg = ascii_letters
        port.write(msg.encode())
        buf = port.read(len(ascii_letters)).decode()
        self.assertEqual(msg, buf)
        port.close()

    def test_baudrate_fs_dev(self):
        """Check baudrate settings for full speed devices."""
        with open('pyftdi/tests/resources/ft230x.yaml', 'rb') as yfp:
            self.loader.load(yfp)
        port = serial_for_url('ftdi:///1', baudrate=1200)
        bus, address, _ = port.usb_path
        vftdi = self.loader.get_virtual_ftdi(bus, address)
        vport = vftdi.get_port(1)
        self.assertRaises(ValueError, setattr, port, 'baudrate', 100)
        self.assertRaises(ValueError, setattr, port, 'baudrate', 3100000)
        for baudrate in (200, 600, 1200, 2400, 4800, 9600, 115200, 230400,
                         460800, 490000, 921600, 1000000, 1200000, 1500000,
                         2000000, 3000000):
            port.baudrate = baudrate
            #print(f'{baudrate} -> {port.ftdi.baudrate} -> {vport.baudrate}')
            self.assertEqual(port.ftdi.baudrate, vport.baudrate)
        port.close()

    def test_baudrate_hs_dev(self):
        """Check baudrate settings for high speed devices."""
        with open('pyftdi/tests/resources/ft232h.yaml', 'rb') as yfp:
            self.loader.load(yfp)
        port = serial_for_url('ftdi:///1', baudrate=1200)
        bus, address, _ = port.usb_path
        vftdi = self.loader.get_virtual_ftdi(bus, address)
        vport = vftdi.get_port(1)
        self.assertRaises(ValueError, setattr, port, 'baudrate', 100)
        self.assertRaises(ValueError, setattr, port, 'baudrate', 12100000)
        for baudrate in (200, 600, 1200, 2400, 4800, 9600, 115200, 230400,
                         460800, 490000, 921600, 1000000, 1200000, 1500000,
                         2000000, 3000000, 4000000, 6000000):
            port.baudrate = baudrate
            #print(f'{baudrate} -> {port.ftdi.baudrate} -> {vport.baudrate}')
            self.assertEqual(port.ftdi.baudrate, vport.baudrate)
        port.close()


class MockRawExtEepromTestCase(FtdiTestCase):
    """Test FTDI EEPROM low-level APIs with external EEPROM device
    """

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        with open('pyftdi/tests/resources/ft232h.yaml', 'rb') as yfp:
            cls.loader.load(yfp)

    def _restore_eeprom(self, ftdi):
        bus, address, _ = ftdi.usb_path
        vftdi = self.loader.get_virtual_ftdi(bus, address)
        data = self.loader.eeprom_backup
        size = len(vftdi.eeprom)
        if len(data) < size:
            data = bytearray(data) + bytes(size-len(data))
        vftdi.eeprom = bytes(data)

    def test_dump(self):
        """Check EEPROM full content."""
        ftdi = Ftdi()
        ftdi.open_from_url('ftdi:///1')
        self._restore_eeprom(ftdi)
        ref_data = bytes(list(range(256)))
        size = len(ref_data)
        data = ftdi.read_eeprom()
        self.assertEqual(len(data), size)
        self.assertEqual(ref_data, data)
        ftdi.close()

    def test_random_access_read(self):
        """Check EEPROM random read access."""
        ftdi = Ftdi()
        ftdi.open_from_url('ftdi:///1')
        self._restore_eeprom(ftdi)
        ref_data = bytes(list(range(256)))
        size = len(ref_data)
        # out of bound
        self.assertRaises(ValueError, ftdi.read_eeprom, size, 2)
        # last bytes
        buf = ftdi.read_eeprom(size-2, 2)
        self.assertEqual(buf[0:2], ref_data[-2:])
        self.assertEqual(buf[0:2], b'\xfe\xff')
        # out of bound
        self.assertRaises(ValueError, ftdi.read_eeprom, size-2, 4)
        # unaligned access
        buf = ftdi.read_eeprom(1, 2)
        self.assertEqual(buf[0:2], ref_data[1:3])
        self.assertEqual(buf[0:2], b'\x01\x02')
        # long read, unaligned access, unaligned size
        buf = ftdi.read_eeprom(43, 43)
        self.assertEqual(len(buf), 43)
        self.assertEqual(buf, ref_data[43:86])
        ftdi.close()

    def test_randow_access_write(self):
        """Check EEPROM random write access."""
        ftdi = Ftdi()
        ftdi.open_from_url('ftdi:///1')
        bus, address, _ = ftdi.usb_path
        vftdi = self.loader.get_virtual_ftdi(bus, address)
        self._restore_eeprom(ftdi)
        checksum1 = vftdi.eeprom[-2:]
        orig_data = vftdi.eeprom[:8]
        ref_data = b'ABCD'
        ftdi.write_eeprom(0, ref_data, dry_run=False)
        checksum2 = vftdi.eeprom[-2:]
        # verify the data have been written
        self.assertEqual(vftdi.eeprom[:4], ref_data)
        # verify the data have not been overwritten
        self.assertEqual(vftdi.eeprom[4:8], orig_data[4:])
        # verify the checksum has been updated
        # TODO compute the expected checksum
        self.assertNotEqual(checksum1, checksum2)
        checksum1 = vftdi.eeprom[-2:]
        orig_data = vftdi.eeprom[:24]
        ftdi.write_eeprom(9, ref_data, dry_run=False)
        checksum2 = vftdi.eeprom[-2:]
        # verify the unaligned data have been written
        self.assertEqual(vftdi.eeprom[9:13], ref_data)
        # verify the data have not been overwritten
        self.assertEqual(vftdi.eeprom[:9], orig_data[:9])
        self.assertEqual(vftdi.eeprom[13:24], orig_data[13:])
        # verify the checksum has been updated
        self.assertNotEqual(checksum1, checksum2)
        checksum1 = vftdi.eeprom[-2:]
        orig_data = vftdi.eeprom[:48]
        ftdi.write_eeprom(33, ref_data[:3], dry_run=False)
        checksum2 = vftdi.eeprom[-2:]
        # verify the unaligned data have been written
        self.assertEqual(vftdi.eeprom[33:36], ref_data[:3])
        # verify the data have not been overwritten
        self.assertEqual(vftdi.eeprom[:33], orig_data[:33])
        self.assertEqual(vftdi.eeprom[36:48], orig_data[36:])
        # verify the checksum has been updated
        self.assertNotEqual(checksum1, checksum2)


class MockRawIntEepromTestCase(FtdiTestCase):
    """Test FTDI EEPROM low-level APIs with internal EEPROM device
    """

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        with open('pyftdi/tests/resources/ft230x.yaml', 'rb') as yfp:
            cls.loader.load(yfp)

    def test_descriptor_update(self):
        """Check EEPROM content overrides YaML configuration."""
        # this test is more about testing the virtual FTDI infrastructure
        # than a pure PyFtdi test
        devs = Ftdi.list_devices('ftdi:///?')
        self.assertEqual(len(devs), 1)
        desc = devs[0][0]
        # these values are not the ones defined in YaML, but stored in EEPROM
        self.assertEqual(desc.sn, 'FT3KMGTL')
        self.assertEqual(desc.description, 'LC231X')

    def test_eeprom_read(self):
        """Check full read sequence."""
        ftdi = Ftdi()
        ftdi.open_from_url('ftdi:///1')
        data = ftdi.read_eeprom()
        self.assertEqual(len(data), 0x400)
        ftdi.close()


class MockCBusEepromTestCase(FtdiTestCase):
    """Test FTDI EEPROM APIs that manage CBUS feature
    """

    def tearDown(self):
        self.loader.unload()

    def test_ft230x(self):
        with open('pyftdi/tests/resources/ft230x.yaml', 'rb') as yfp:
            self.loader.load(yfp)
        eeprom = FtdiEeprom()
        eeprom.open('ftdi:///1')
        # default EEPROM config does not have any CBUS configured as GPIO
        self.assertEqual(eeprom.cbus_pins, [])
        self.assertEqual(eeprom.cbus_mask, 0)
        # enable CBUS1 and CBUS3 as GPIO
        eeprom.set_property('cbus_func_1', 'gpio')
        eeprom.set_property('cbus_func_3', 'gpio')
        eeprom.sync()
        self.assertEqual(eeprom.cbus_pins, [1, 3])
        self.assertEqual(eeprom.cbus_mask, 0xA)
        # enable CBUS0 and CBUS2 as GPIO
        eeprom.set_property('cbus_func_0', 'gpio')
        eeprom.set_property('cbus_func_2', 'gpio')
        # not yet committed
        self.assertEqual(eeprom.cbus_pins, [1, 3])
        self.assertEqual(eeprom.cbus_mask, 0xA)
        eeprom.sync()
        # committed
        self.assertEqual(eeprom.cbus_pins, [0, 1, 2, 3])
        self.assertEqual(eeprom.cbus_mask, 0xF)
        # invalid CBUS pin
        self.assertRaises(ValueError, eeprom.set_property,
                          'cbus_func_4', 'gpio')
        # invalid pin function
        self.assertRaises(ValueError, eeprom.set_property,
                          'cbus_func_0', 'gpio_')
        # invalid pin
        self.assertRaises(ValueError, eeprom.set_property,
                          'cbus_func', 'gpio')
        # valid alternative mode
        eeprom.set_property('cbus_func_0', 'txled')
        eeprom.set_property('cbus_func_1', 'rxled')
        eeprom.sync()
        self.assertEqual(eeprom.cbus_pins, [2, 3])
        self.assertEqual(eeprom.cbus_mask, 0xC)
        eeprom.close()

    def test_ft232h(self):
        with open('pyftdi/tests/resources/ft232h_x2.yaml', 'rb') as yfp:
            self.loader.load(yfp)
        eeprom = FtdiEeprom()
        eeprom.open('ftdi://::FT1ABC1/1', ignore=True)
        eeprom.erase()
        eeprom.initialize()
        # default EEPROM config does not have any CBUS configured as GPIO
        self.assertEqual(eeprom.cbus_pins, [])
        self.assertEqual(eeprom.cbus_mask, 0)
        eeprom.set_property('cbus_func_6', 'gpio')
        eeprom.set_property('cbus_func_9', 'gpio')
        # not yet committed
        self.assertEqual(eeprom.cbus_pins, [])
        self.assertEqual(eeprom.cbus_mask, 0)
        eeprom.sync()
        # committed
        self.assertEqual(eeprom.cbus_pins, [6, 9])
        self.assertEqual(eeprom.cbus_mask, 0xA)
        eeprom.set_property('cbus_func_5', 'gpio')
        eeprom.set_property('cbus_func_8', 'gpio')
        eeprom.sync()
        self.assertEqual(eeprom.cbus_pins, [5, 6, 8, 9])
        self.assertEqual(eeprom.cbus_mask, 0xF)
        # pin1 and pin3 is not configurable as GPIO
        self.assertRaises(ValueError, eeprom.set_property,
                          'cbus_func_1', 'gpio')
        self.assertRaises(ValueError, eeprom.set_property,
                          'cbus_func_3', 'gpio')
        eeprom.close()


class MockCbusGpioTestCase(FtdiTestCase):
    """Test FTDI CBUS GPIO APIs
    """

    def tearDown(self):
        self.loader.unload()

    def test_230x(self):
        """Check simple GPIO write and read sequence."""
        # load custom CBUS config, with:
            # CBUS0: GPIO (gpio)
            # CBUS1: GPIO (gpio)
            # CBUS0: DRIVE1 (forced to high level)
            # CBUS0: TXLED  (eq. to highz for tests)
        with open('pyftdi/tests/resources/ft230x_io.yaml', 'rb') as yfp:
            self.loader.load(yfp)
        ftdi = Ftdi()
        ftdi.open_from_url('ftdi:///1')
        self.assertEqual(ftdi.has_cbus, True)
        vftdi = self.loader.get_virtual_ftdi(1, 1)
        vport = vftdi.get_port(1)
        # CBUS0: in, CBUS1: out, CBUS2: in, CBUS3: out
        #   however, only CBUS0 and CBUS1 are mapped as GPIO,
        #   CBUS2 forced to 1 and CBUS3 not usable as IO
        #   even if use mask is 1111
        eeprom_mask = 0b0011
        eeprom_force = 0b0100
        cbus_mask = 0b1111
        cbus_dir = 0b1010
        ftdi.set_cbus_direction(cbus_mask, cbus_dir)
        cbus_out = 0b0011
        # CBUS0: 1, CBUS1: 1
        #   however, only CBUS1 is out, so CBUS0 output value should be ignored
        ftdi.set_cbus_gpio(cbus_out)
        exp_out = cbus_dir & cbus_out
        exp_out &= eeprom_mask
        exp_out |= eeprom_force
        vcbus, vactive = vport.cbus
        self.assertEqual(vcbus, exp_out)
        self.assertEqual(vactive, eeprom_mask | eeprom_force)
        cbus_out = 0b0000
        ftdi.set_cbus_gpio(cbus_out)
        exp_out = cbus_dir & cbus_out
        exp_out &= eeprom_mask
        exp_out |= eeprom_force
        vcbus, vactive = vport.cbus
        self.assertEqual(vcbus, exp_out)
        cbus_in = 0b0101
        vport.cbus = cbus_in
        cbus = ftdi.get_cbus_gpio()
        exp_in = cbus_in & eeprom_mask
        self.assertEqual(cbus, exp_in)
        ftdi.close()

    def test_lc231x(self):
        """Check simple GPIO write and read sequence."""
         # load custom CBUS config, with:
            # CBUS0: GPIO (gpio)
            # CBUS1: TXLED
            # CBUS2: DRIVE0 (to light up RX green led)
            # CBUS3: GPIO (gpio)
            # only CBUS0 and CBUS3 are available on LC231X
            # CBUS1 is connected to TX led, CBUS2 to RX led
        with open('pyftdi/tests/resources/ft231x_cbus.yaml', 'rb') as yfp:
            self.loader.load(yfp)
        ftdi = Ftdi()
        ftdi.open_from_url('ftdi:///1')
        self.assertEqual(ftdi.has_cbus, True)
        vftdi = self.loader.get_virtual_ftdi(1, 1)
        vport = vftdi.get_port(1)
        # CBUS0: in, CBUS1: out, CBUS2: in, CBUS3: out
        #   however, only CBUS0 and CBUS3 are mapped as GPIO,
        #   CBUS1 not usable as IO, CBUS2 is fixed to low
        #   even if use mask is 1111
        eeprom_mask = 0b1001
        eeprom_force_low = 0b0100
        cbus_mask = 0b1111
        cbus_dir = 0b1010
        ftdi.set_cbus_direction(cbus_mask, cbus_dir)
        cbus_out = 0b1111
        # however, only CBUS0 & 3 are out, so CBUS1/CBUS2 should be ignored
        ftdi.set_cbus_gpio(cbus_out)
        exp_out = cbus_dir & cbus_out
        exp_out &= eeprom_mask
        vcbus, vactive = vport.cbus
        self.assertEqual(vcbus, exp_out)
        self.assertEqual(vactive, eeprom_mask | eeprom_force_low)
        cbus_out = 0b0000
        ftdi.set_cbus_gpio(cbus_out)
        exp_out = cbus_dir & cbus_out
        exp_out &= eeprom_mask
        vcbus, vactive = vport.cbus
        self.assertEqual(vcbus, exp_out)
        cbus_in = 0b0101
        vport.cbus = cbus_in
        cbus = ftdi.get_cbus_gpio()
        exp_in = cbus_in & eeprom_mask
        self.assertEqual(cbus, exp_in)
        ftdi.close()


def suite():
    suite_ = TestSuite()
    suite_.addTest(makeSuite(MockUsbToolsTestCase, 'test'))
    suite_.addTest(makeSuite(MockFtdiDiscoveryTestCase, 'test'))
    suite_.addTest(makeSuite(MockSimpleDeviceTestCase, 'test'))
    suite_.addTest(makeSuite(MockDualDeviceTestCase, 'test'))
    suite_.addTest(makeSuite(MockTwoPortDeviceTestCase, 'test'))
    suite_.addTest(makeSuite(MockFourPortDeviceTestCase, 'test'))
    suite_.addTest(makeSuite(MockManyDevicesTestCase, 'test'))
    suite_.addTest(makeSuite(MockSimpleDirectTestCase, 'test'))
    suite_.addTest(makeSuite(MockSimpleMpsseTestCase, 'test'))
    suite_.addTest(makeSuite(MockSimpleGpioTestCase, 'test'))
    suite_.addTest(makeSuite(MockSimpleUartTestCase, 'test'))
    suite_.addTest(makeSuite(MockRawExtEepromTestCase, 'test'))
    suite_.addTest(makeSuite(MockRawIntEepromTestCase, 'test'))
    suite_.addTest(makeSuite(MockCBusEepromTestCase, 'test'))
    suite_.addTest(makeSuite(MockCbusGpioTestCase, 'test'))
    return suite_


def main():
    testmod(modules[__name__])
    debug = to_bool(environ.get('FTDI_DEBUG', 'off'))
    if debug:
        formatter = logging.Formatter('%(asctime)s.%(msecs)03d %(levelname)-7s'
                                      ' %(name)-18s [%(lineno)4d] %(message)s',
                                      '%H:%M:%S')
    else:
        formatter = logging.Formatter('%(message)s')
    level = environ.get('FTDI_LOGLEVEL', 'warning').upper()
    try:
        loglevel = getattr(logging, level)
    except AttributeError as exc:
        raise ValueError(f'Invalid log level: {level}') from exc
    FtdiLogger.log.addHandler(logging.StreamHandler(stdout))
    FtdiLogger.set_level(loglevel)
    FtdiLogger.set_formatter(formatter)
    # Force PyUSB to use PyFtdi test framework for USB backends
    UsbTools.BACKENDS = ('backend.usbvirt', )
    # Ensure the virtual backend can be found and is loaded
    backend = UsbTools.find_backend()
    try:
        # obtain the loader class associated with the virtual backend
        global MockLoader
        MockLoader = backend.create_loader()
    except AttributeError as exc:
        raise AssertionError('Cannot load virtual USB backend') from exc
    ut_main(defaultTest='suite')


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2017-2020, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

import logging
from unittest import TestCase, TestSuite, main as ut_main, makeSuite
from binascii import hexlify
from doctest import testmod
from os import environ
from sys import modules, stdout
from pyftdi import FtdiLogger
from pyftdi.i2c import I2cController, I2cIOError
from pyftdi.misc import pretty_size

#pylint: disable-msg=attribute-defined-outside-init
#pylint: disable-msg=missing-docstring
#pylint: disable-msg=no-self-use


class I2cTca9555TestCase(TestCase):
    """Simple test for a TCA9555 device on I2C bus @ address 0x21
    """

    def test(self):
        self._i2c = I2cController()
        self._open()
        self._read_it()
        self._write_it()
        self._close()

    def _open(self):
        """Open an I2c connection to a slave"""
        url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        self._i2c.configure(url)

    def _read_it(self):
        port = self._i2c.get_port(0x21)
        port.exchange([0x04], 1)

    def _write_it(self):
        port = self._i2c.get_port(0x21)
        port.write_to(0x06, b'\x00')
        port.write_to(0x02, b'\x55')
        port.read_from(0x00, 1)

    def _close(self):
        """Close the I2C connection"""
        self._i2c.terminate()


class I2cAccelTest(TestCase):
    """Basic test for an ADXL345 device on I2C bus @ address 0x53
    """

    def test(self):
        self._i2c = I2cController()
        self._open()
        self._read_device_id()
        self._close()

    def _open(self):
        """Open an I2c connection to a slave"""
        url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        self._i2c.configure(url)

    def _read_device_id(self):
        port = self._i2c.get_port(0x53)
        device_id = port.exchange([0x00], 1)
        hex_device_id = hexlify(device_id).decode()
        print('DEVICE ID:', hex_device_id)
        return hex_device_id

    def _close(self):
        """Close the I2C connection"""
        self._i2c.terminate()


class I2cReadTest(TestCase):
    """Simple test to read a sequence of bytes I2C bus @ address 0x36
    """

    def test(self):
        self._i2c = I2cController()
        self._open()
        self._read()
        self._close()

    def _open(self):
        """Open an I2c connection to a slave"""
        url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        self._i2c.configure(url)

    def _read(self):
        address = environ.get('I2C_ADDRESS', '0x36').lower()
        addr = int(address, 16 if address.startswith('0x') else 10)
        port = self._i2c.get_port(addr)
        data = port.read(32)
        print(hexlify(data).decode(), data.decode('utf8', errors='replace'))

    def _close(self):
        """Close the I2C connection"""
        self._i2c.terminate()


class I2cEepromTest(TestCase):
    """Simple test to read a sequence of bytes I2C bus @ address 0x50,
       from an I2C data flash
    """

    @classmethod
    def setUpClass(cls):
        cls.url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        address = environ.get('I2C_ADDRESS', '0x50').lower()
        cls.address = int(address, 16 if address.startswith('0x') else 10)

    def setUp(self):
        self._i2c = I2cController()
        self._i2c.configure(self.url, frequency=400e3,
                            clockstretching=False, debug=False, rdoptim=True)

    def tearDown(self):
        self._i2c.terminate()

    def test_short(self):
        port = self._i2c.get_port(self.address)
        # select start address
        port.write(b'\x00\x08')
        data = port.read(4)
        text = data.decode('utf8', errors='replace')
        # print(hexlify(data).decode(), text)
        self.assertEqual(text, 'Worl')

    def test_long(self):
        port = self._i2c.get_port(self.address)
        # select start address
        #print('RC', self._i2c.ftdi.read_data_get_chunksize())
        #print('WC', self._i2c.ftdi.write_data_get_chunksize())
        from time import time as now
        size = 4096
        port.write(b'\x00\x00')
        start = now()
        data = port.read(size)
        stop = now()
        text = data.decode('utf8', errors='replace')
        delta = stop-start
        byterate = pretty_size(len(data)/delta)
        print(f'Exec time: {1000*delta:.3f} ms, {byterate}/s')
        self.assertEqual(text[8:12], 'Worl')


class I2cReadGpioTest(TestCase):
    """Simple test to exercise I2C + GPIO mode.

       A slave device (such as EEPROM) should be connected to the I2C bus
       at either the default 0x36 address or defined with the I2C_ADDRESS
       ebvironment variable.

       AD0: SCL, AD1+AD2: SDA, AD3 connected to AC0, AD4 connected to AC1

       I2C EEPROM is read, and values are written to AD3:AD4 and read back
       from AC0:AC1.
    """

    GPIO_WIDTH = 2  # use 2 GPIOs for output, 2 GPIOs for input (loopback)
    GPIO_OUT_OFFSET = 3  # GPIO output are b3..b4
    GPIO_IN_OFFSET = 8  # GPIO input are b8..b9

    def test(self):
        self._i2c = I2cController()
        self._open()
        self._execute_sequence()
        self._execute_interleave()
        self._close()

    def _open(self):
        """Open an I2c connection to a slave"""
        url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        self._i2c.configure(url)
        address = environ.get('I2C_ADDRESS', '0x36').lower()
        addr = int(address, 16 if address.startswith('0x') else 10)
        self._port = self._i2c.get_port(addr)
        self._gpio = self._i2c.get_gpio()
        mask = (1 << self.GPIO_WIDTH) - 1
        self._gpio.set_direction(mask << self.GPIO_OUT_OFFSET |
                                 mask << self.GPIO_IN_OFFSET,
                                 mask << self.GPIO_OUT_OFFSET)

    def _execute_sequence(self):
        # reset EEPROM read pointer position
        self._port.write(b'\x00\x00')
        ref = self._port.read(32)
        for dout in range(1 << self.GPIO_WIDTH):
            self._gpio.write(dout << self.GPIO_OUT_OFFSET)
            din = self._gpio.read() >> self.GPIO_IN_OFFSET
            if dout != din:
                raise AssertionError(f'GPIO mismatch {din:04x} != {dout:04x}')
        self._gpio.write(0)
        # reset EEPROM read pointer position
        self._port.write(b'\x00\x00')
        data = self._port.read(32)
        if data != ref:
            raise AssertionError("I2C data mismatch")

    def _execute_interleave(self):
        # reset EEPROM read pointer position
        self._port.write(b'\x00\x00')
        ref = self._port.read(32)
        for dout in range(1 << self.GPIO_WIDTH):
            self._gpio.write(dout << self.GPIO_OUT_OFFSET)
            # reset EEPROM read pointer position
            self._port.write(b'\x00\x00')
            data = self._port.read(32)
            din = self._gpio.read() >> self.GPIO_IN_OFFSET
            if data != ref:
                raise AssertionError("I2C data mismatch")
            if dout != din:
                raise AssertionError(f'GPIO mismatch {din:04x} != {dout:04x}')
        self._gpio.write(0)

    def _close(self):
        """Close the I2C connection"""
        self._i2c.terminate()


class I2cClockStrechingGpioCheck(TestCase):
    """Simple test to check clock stretching cannot be overwritten with
       GPIOs.
    """

    def test(self):
        self._i2c = I2cController()
        url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        self._i2c.configure(url, clockstretching=True)
        gpio = self._i2c.get_gpio()
        self.assertRaises(I2cIOError, gpio.set_direction, 1 << 7, 0)


class I2cDualMaster(TestCase):
    """Check the behaviour of 2 I2C masters. Requires a multi port FTDI device,
       i.e. FT2232H or FT4232H. See issue #159.
    """

    def test(self):
        url1 = environ.get('FTDI_DEVICE', 'ftdi:///1')
        i2c1 = I2cController()
        i2c1.configure(url1, frequency=100000)
        url2 = '%s%d' % (url1[:-1], int(url1[-1])+1)
        i2c2 = I2cController()
        i2c2.configure(url2, frequency=100000)
        port = i2c2.get_port(0x76)
        print(port.read_from(0x00, 2))
        print(port.read_from(0x00, 2))


class I2cIssue143(TestCase):
    """#143.
    """

    def test(self):
        url = environ.get('FTDI_DEVICE', 'ftdi:///1')
        address = environ.get('I2C_ADDRESS', '0x50').lower()
        addr = int(address, 16 if address.startswith('0x') else 10)
        i2c = I2cController()
        i2c.configure(url)
        slave = i2c.get_port(addr)
        gpio = i2c.get_gpio()
        gpio.set_direction(0x0010, 0x0010)
        gpio.write(0)
        gpio.write(1<<4)
        gpio.write(0)
        slave.write([0x12, 0x34])
        gpio.write(0)
        gpio.write(1<<4)
        gpio.write(0)


def suite():
    """FTDI I2C driver test suite

       Simple test to demonstrate I2C.

       Please ensure that the HW you connect to the FTDI port A does match
       the encoded configuration. GPIOs can be driven high or low, so check
       your HW setup before running this test as it might damage your HW.

       Do NOT run this test if you use FTDI port A as an UART or SPI
       bridge -or any unsupported setup!! You've been warned.
    """
    ste = TestSuite()
    #ste.addTest(I2cTca9555TestCase('test'))
    #ste.addTest(I2cAccelTest('test'))
    #ste.addTest(I2cReadTest('test'))
    ste.addTest(makeSuite(I2cEepromTest, 'test'))
    #ste.addTest(I2cReadGpioTest('test'))
    ste.addTest(I2cClockStrechingGpioCheck('test'))
    #ste.addTest(I2cDualMaster('test'))
    ste.addTest(I2cIssue143('test'))
    return ste


def main():
    testmod(modules[__name__])
    FtdiLogger.log.addHandler(logging.StreamHandler(stdout))
    level = environ.get('FTDI_LOGLEVEL', 'info').upper()
    try:
        loglevel = getattr(logging, level)
    except AttributeError as exc:
        raise ValueError(f'Invalid log level: {level}') from exc
    FtdiLogger.set_level(loglevel)
    ut_main(defaultTest='suite')


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""Quick and dirty bytearray vs. array('B') performance test."""

from array import array
from struct import pack
from timeit import timeit
from time import perf_counter

def timing(f, n, a):
    start = perf_counter()
    for _ in range(n):
        f(a); f(a); f(a); f(a); f(a); f(a); f(a); f(a); f(a); f(a)
    finish = perf_counter()
    return '%s\t%f' % (f.__name__, finish - start)

def time_array(addr):
    return array('B', addr)

def time_bytearray(addr):
    return bytearray(addr)

def extend_array(addr):
    b = bytearray()
    b.extend(addr)
    b.extend(b)
    b.extend(b)
    b.extend(b)
    b.extend(b)
    b.extend(b)
    return b

def extend_bytearray(addr):
    b = bytearray()
    b.extend(addr)
    b.extend(b)
    b.extend(b)
    b.extend(b)
    b.extend(b)
    b.extend(b)
    return b

def array_tostring(addr):
    return array('B', addr).tobytes()

def str_bytearray(addr):
    return str(bytearray(addr))

def struct_pack(addr):
    return pack('4B', *addr)

def main():
    count = 100000
    addr = '192.168.4.2'
    addr = tuple([int(i) for i in addr.split('.')])
    print('\t\ttiming\t\tfunc\t\tno func')
    print('%s\t%s\t%s' % (timing(time_array, count, addr),
          timeit('time_array((192,168,4,2))', number=count,
                 setup='from __main__ import time_array'),
          timeit("array('B', (192,168,4,2))", number=count,
                 setup='from array import array')))
    print('%s\t%s\t%s' % (timing(time_bytearray, count, addr),
          timeit('time_bytearray((192,168,4,2))', number=count,
                 setup='from __main__ import time_bytearray'),
          timeit('bytearray((192,168,4,2))', number=count)))
    print('%s\t%s' % (timing(extend_array, count, addr),
          timeit('extend_array((192,168,4,2))', number=count,
                 setup='from __main__ import extend_array')))
    print('%s\t%s' % (timing(extend_bytearray, count, addr),
          timeit('extend_bytearray((192,168,4,2))', number=count,
                 setup='from __main__ import extend_bytearray')))
    print('%s\t%s\t%s' % (timing(array_tostring, count, addr),
          timeit('array_tostring((192,168,4,2))', number=count,
                 setup='from __main__ import array_tostring'),
          timeit("array('B', (192,168,4,2)).tostring()", number=count,
                 setup='from array import array')))
    print('%s\t%s\t%s' % (timing(str_bytearray, count, addr),
          timeit('str_bytearray((192,168,4,2))', number=count,
                 setup='from __main__ import str_bytearray'),
          timeit('str(bytearray((192,168,4,2)))', number=count)))
    print('%s\t%s\t%s' % (timing(struct_pack, count, addr),
          timeit('struct_pack((192,168,4,2))', number=count,
                 setup='from __main__ import struct_pack'),
          timeit("pack('4B', *(192,168,4,2))", number=count,
                 setup='from struct import pack')))

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2020, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.

#pylint: disable-msg=empty-docstring
#pylint: disable-msg=missing-docstring
#pylint: disable-msg=no-self-use
#pylint: disable-msg=invalid-name
#pylint: disable-msg=global-statement

from doctest import testmod
from importlib import import_module
from os.path import dirname, join as joinpath
from sys import modules, path as syspath
from unittest import TestCase, TestSuite, makeSuite, main as ut_main


class ToolsTestCase(TestCase):
    """Test tool suite can be loaded.

       This is especially useful to find Python syntax version mismatch
       and other not-yet-supported modules/features.

       PyFtdi and tools should support Python 3.7 onwards.
    """

    @classmethod
    def setUpClass(cls):
        tools_path = joinpath(dirname(dirname(__file__)), 'bin')
        syspath.append(tools_path)

    def test_ftconf(self):
        """Test ftconf.py tool"""
        mod = import_module('ftconf')
        self.assertIsNot(getattr(mod, 'main', None), mod)
        self.assertIsNot(mod.__doc__, None)

    def test_i2cscan(self):
        """Test ftconf.py tool"""
        mod = import_module('i2cscan')
        self.assertIsNot(getattr(mod, 'main', None), mod)
        self.assertIsNot(mod.__doc__, None)

    def test_pyterm(self):
        """Test ftconf.py tool"""
        mod = import_module('pyterm')
        self.assertIsNot(getattr(mod, 'main', None), mod)
        self.assertIsNot(mod.__doc__, None)

    def test_ftdi_urls(self):
        """Test ftconf.py tool"""
        mod = import_module('ftdi_urls')
        self.assertIsNot(getattr(mod, 'main', None), mod)
        self.assertIsNot(mod.__doc__, None)


def suite():
    suite_ = TestSuite()
    suite_.addTest(makeSuite(ToolsTestCase, 'test'))
    return suite_


def main():
    testmod(modules[__name__])
    ut_main(defaultTest='suite')


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2020, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

import sys
from doctest import testmod
from os import environ
from unittest import TestCase, TestSuite, makeSuite, main as ut_main
from pyftdi.ftdi import Ftdi, FtdiError
from pyftdi.eeprom import FtdiEeprom

#pylint: disable-msg=empty-docstring
#pylint: disable-msg=missing-docstring


class CbusGpioTestCase(TestCase):
    """FTDI CBUS GPIO feature test case"""

    @classmethod
    def setUpClass(cls):
        """Default values"""
        cls.url = environ.get('FTDI_DEVICE', 'ftdi:///1')

    def test_output_gpio(self):
        """Simple test to demonstrate ouput bit-banging on CBUS.

           You need a CBUS-capable FTDI (FT232R/FT232H/FT230X/FT231X), whose
           EEPROM has been configured to support GPIOs on CBUS0 and CBUS3.

           Hard-wiring is required to run this test:
           * CBUS0 (output) should be connected to CTS (input)
           * CBUS3 (output) should be connected to DSR (input)
        """
        ftdi = Ftdi()
        ftdi.open_from_url(self.url)
        # sanity check: device should support CBUS feature
        self.assertEqual(ftdi.has_cbus, True)
        eeprom = FtdiEeprom()
        eeprom.connect(ftdi)
        # sanity check: device should have been configured for CBUS GPIOs
        self.assertEqual(eeprom.cbus_mask & 0b1001, 0b1001)
        # configure CBUS0 and CBUS3 as output
        ftdi.set_cbus_direction(0b1001, 0b1001)
        # no input pin available
        self.assertRaises(FtdiError, ftdi.get_cbus_gpio)
        for cycle in range(40):
            value = cycle & 0x3
            # CBUS0 and CBUS3
            cbus = ((value & 0x2) << 2) | value & 0x1
            # for now, need a digital/logic analyzer to validate output
            ftdi.set_cbus_gpio(cbus)
            # CBUS0 is connected to CTS, CBUS3 to DSR
            # need to inverse logical level as RS232 uses negative logic
            sig = int(not ftdi.get_cts()) | (int(not ftdi.get_dsr()) << 1)
            self.assertEqual(value, sig)

    def test_input_gpio(self):
        """Simple test to demonstrate input bit-banging on CBUS.

           You need a CBUS-capable FTDI (FT232R/FT232H/FT230X/FT231X), whose
           EEPROM has been configured to support GPIOs on CBUS0 and CBUS3.

           Hard-wiring is required to run this test:
           * CBUS0 (input) should be connected to RTS (output)
           * CBUS3 (input) should be connected to DTR (output)
        """
        ftdi = Ftdi()
        ftdi.open_from_url(self.url)
        # sanity check: device should support CBUS feature
        self.assertEqual(ftdi.has_cbus, True)
        eeprom = FtdiEeprom()
        eeprom.connect(ftdi)
        # sanity check: device should have been configured for CBUS GPIOs
        self.assertEqual(eeprom.cbus_mask & 0b1001, 0b1001)
        # configure CBUS0 and CBUS3 as input
        ftdi.set_cbus_direction(0b1001, 0b0000)
        # no output pin available
        self.assertRaises(FtdiError, ftdi.set_cbus_gpio, 0)
        for cycle in range(40):
            rts = bool(cycle & 0x1)
            dtr = bool(cycle & 0x2)
            ftdi.set_rts(rts)
            ftdi.set_dtr(dtr)
            # need to inverse logical level as RS232 uses negative logic
            cbus = ~ftdi.get_cbus_gpio()
            sig = (cbus & 0x1) | ((cbus & 0x8) >> 2)
            value = cycle & 0x3
            self.assertEqual(value, sig)


def suite():
    suite_ = TestSuite()
    # peak the test that matches your HW setup, see test doc for details
    # suite_.addTest(makeSuite(CbusGpioTestCase, 'test_output'))
    suite_.addTest(makeSuite(CbusGpioTestCase, 'test_input'))
    return suite_


def main():
    testmod(sys.modules[__name__])
    ut_main(defaultTest='suite')


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (c) 2017-2020, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

import logging
from doctest import testmod
from os import environ
from multiprocessing import Process, set_start_method
from random import choice, seed
from string import printable
from struct import calcsize as scalc, pack as spack, unpack as sunpack
from sys import modules, platform, stdout
from time import sleep, time as now
from threading import Thread
from unittest import TestCase, TestSuite, skipIf, makeSuite, main as ut_main
from pyftdi import FtdiLogger
from pyftdi.ftdi import Ftdi
from pyftdi.misc import to_bool
from pyftdi.serialext import serial_for_url

#pylint: disable-msg=missing-docstring
#pylint: disable-msg=protected-access

# Specify the second port for multi port device
# Unfortunately, auto detection triggers some issue in multiprocess test
URL_BASE = environ.get('FTDI_DEVICE', 'ftdi:///2')
URL = ''.join((URL_BASE[:-1], '1'))
IFCOUNT = int(URL_BASE[-1])


class FtdiTestCase(TestCase):
    """Common features for all tests.
    """

    @classmethod
    def setUpClass(cls):
        cls.debug = to_bool(environ.get('FTDI_DEBUG', 'off'),
                                 permissive=False)

    def setUp(self):
        if self.debug:
            print('.'.join(self.id().split('.')[-2:]))


class UartTestCase(FtdiTestCase):
    """FTDI UART driver test case

       Simple test to demonstrate UART feature.

       Depending on your FTDI device, you need to either:
         * connect RXD and TXD on the unique port for 1-port FTDI device
           (AD0 and AD1)
         * connect RXD1 and TXD2, and RXD2 and TXD1 on the two first port of
           2- or 4- port FTDI devices (AD0 - BD1 and AD1 - BD0)

       Do NOT run this test if you use FTDI port(s) as an SPI or I2C
       bridge -or any unsupported setup!! You've been warned.
    """

    COUNT = 512

    @classmethod
    def setUpClass(cls):
        FtdiTestCase.setUpClass()
        seed()

    @skipIf(IFCOUNT < 2, 'Device has not enough UART interfaces')
    def test_uart_cross_talk_sp(self):
        """Exchange a random byte stream between the two first UART interfaces
           of the same FTDI device, from the same process

           This also validates PyFtdi support to use several interfaces on the
           same FTDI device from the same Python process
        """
        something_out = self.generate_bytes()
        urla = URL
        urlb = self.build_next_url(urla)
        porta = serial_for_url(urla, baudrate=1000000)
        portb = serial_for_url(urlb, baudrate=1000000)
        try:
            if not porta.is_open:
                porta.open()
            if not portb.is_open:
                portb.open()
            # print("porta: %d:%d:%d" % porta.usb_path)
            # print("portb: %d:%d:%d" % portb.usb_path)
            porta.timeout = 1.0
            portb.timeout = 1.0
            something_out = self.generate_bytes()
            porta.write(something_out)
            something_in = portb.read(len(something_out))
            self.assertEqual(len(something_in), len(something_out))
            self.assertEqual(something_in, something_out)
            something_out = self.generate_bytes()
            portb.write(something_out)
            something_in = porta.read(len(something_out))
            self.assertEqual(len(something_in), len(something_out))
            self.assertEqual(something_in, something_out)
        finally:
            porta.close()
            portb.close()

    @skipIf(IFCOUNT < 2, 'Device has not enough UART interfaces')
    @skipIf(platform == 'win32', 'Not tested on Windows')
    def test_uart_cross_talk_mp(self):
        if IFCOUNT > 4:
            raise IOError('No FTDI device')
        urla = URL
        urlb = self.build_next_url(urla)
        something_out = self.generate_bytes()
        proca = Process(target=self._cross_talk_write_then_read,
                        args=(urla, something_out))
        procb = Process(target=self._cross_talk_read_then_write,
                        args=(urlb, something_out))
        # start B first to ensure RX port is open before TX port starts
        # emitting
        procb.start()
        sleep(0.25)
        proca.start()
        proca.join(2)
        procb.join(2)
        # although the failure is reported (and traceback shown) in the
        # subprocess, we still need to fail the main process test in this case
        exita = proca.exitcode
        exitb = procb.exitcode
        self.assertEqual(exita, 0)
        self.assertEqual(exitb, 0)

    @skipIf(IFCOUNT != 1, 'Test reserved for single-port FTDI device')
    def test_uart_loopback(self):
        """Exchange a random byte stream between the two first UART interfaces
           of the same FTDI device, from the same process

           This also validates PyFtdi support to use several interfaces on the
           same FTDI device from the same Python process
        """
        something_out = self.generate_bytes()
        port = serial_for_url(URL, baudrate=1000000)
        for _ in range(10):
            try:
                if not port.is_open:
                    port.open()
                port.timeout = 1.0
                something_out = self.generate_bytes()
                port.write(something_out)
                something_in = port.read(len(something_out))
                self.assertEqual(len(something_in), len(something_out))
                self.assertEqual(something_in, something_out)
            finally:
                port.close()

    @skipIf(IFCOUNT < 2, 'Device has not enough UART interfaces')
    def test2_uart_cross_talk_speed(self):
        urla = URL
        urlb = self.build_next_url(urla)
        porta = serial_for_url(urla, baudrate=6000000)
        portb = serial_for_url(urlb, baudrate=6000000)
        size = int(1e6)
        results = [None, None]
        chunk = 537
        source = Thread(target=self._stream_source,
                        args=(porta, chunk, size, results),
                        daemon=True)
        sink = Thread(target=self._stream_sink,
                      args=(portb, size, results),
                      daemon=True)
        sink.start()
        source.start()
        source.join()
        sleep(0.5)
        sink.join()
        if isinstance(results[1], Exception):
            #pylint: disable-msg=raising-bad-type
            raise results[1]
        #pylint: disable-msg=unpacking-non-sequence
        tsize, tdelta = results[0]
        rsize, rdelta = results[1]
        self.assertGreater(rsize, 0, 'Not data received')
        if self.debug:
            print(f'TX: {tsize} bytes, {tdelta*1000:.3f} ms, '
                  f'{int(8*tsize/tdelta):d} bps')
            print(f'RX: {rsize} bytes, {rdelta*1000:.3f} ms, '
                  f'{int(8*rsize/rdelta):d} bps')
        self.assertTrue(rsize >= tsize-4*chunk, 'Data loss')

    @skipIf(IFCOUNT != 1, 'Test reserved for single-port FTDI device')
    def test_loopback_talk_speed(self):
        port = serial_for_url(URL, baudrate=6000000)
        size = int(1e6)
        results = [None, None]
        chunk = 537
        source = Thread(target=self._stream_source,
                        args=(port, chunk, size, results),
                        daemon=True)
        sink = Thread(target=self._stream_sink,
                      args=(port, size, results),
                      daemon=True)
        sink.start()
        source.start()
        source.join()
        sleep(0.5)
        sink.join()
        if isinstance(results[1], Exception):
            #pylint: disable-msg=raising-bad-type
            raise results[1]
        #pylint: disable-msg=unpacking-non-sequence
        tsize, tdelta = results[0]
        rsize, rdelta = results[1]
        self.assertGreater(rsize, 0, 'Not data received')
        if self.debug:
            print(f'TX: {tsize} bytes, {tdelta*1000:.3f} ms, '
                  f'{int(8*tsize/tdelta):d} bps')
            print(f'RX: {rsize} bytes, {rdelta*1000:.3f} ms, '
                  f'{int(8*rsize/rdelta):d} bps')
        self.assertTrue(rsize >= tsize-4*chunk, 'Data loss')

    @classmethod
    def _stream_source(cls, port, chunk, size, results):
        pos = 0
        tx_size = 0
        start = now()
        while tx_size < size:
            samples = spack('>%dI' % chunk, *range(pos, pos+chunk))
            pos += chunk
            port.write(samples)
            tx_size += len(samples)
            if results[1] is not None:
                break
        delta = now()-start
        results[0] = tx_size, delta

    @classmethod
    def _stream_sink(cls, port, size, results):
        pos = 0
        first = None
        data = bytearray()
        sample_size = scalc('>I')
        rx_size = 0
        port.timeout = 1.0
        start = now()
        while rx_size < size:
            buf = port.read(1024)
            if not buf:
                print('T')
                break
            rx_size += len(buf)
            data.extend(buf)
            sample_count = len(data)//sample_size
            length = sample_count*sample_size
            samples = sunpack('>%dI' % sample_count, data[:length])
            data = data[length:]
            for sample in samples:
                if first is None:
                    first = sample
                    pos = sample
                    continue
                pos += 1
                if sample != pos:
                    results[1] = AssertionError('Lost byte @ %d', pos-first)
                    return
        delta = now()-start
        results[1] = (rx_size, delta)

    @classmethod
    def _cross_talk_write_then_read(cls, url, refstream):
        # use classmethod & direct AssertionError to avoid pickler issues
        # with multiprocessing:
        # "TypeError: cannot serialize '_io.TextIOWrapper' object"
        port = serial_for_url(url, baudrate=1000000)
        try:
            if not port.is_open:
                port.open()
            port.timeout = 5.0
            port.write(refstream)
            instream = port.read(len(refstream))
            if len(instream) != len(refstream):
                raise AssertionError('Stream length differ')
            # we expect the peer to return the same stream, inverted
            localstream = bytes(reversed(instream))
            if localstream != refstream:
                raise AssertionError('Stream content differ')
        finally:
            port.close()

    @classmethod
    def _cross_talk_read_then_write(cls, url, refstream):
        # use classmethod & direct AssertionError to avoid pickler issues
        # with multiprocessing:
        # "TypeError: cannot serialize '_io.TextIOWrapper' object"
        port = serial_for_url(url, baudrate=1000000)
        try:
            if not port.is_open:
                port.open()
            port.timeout = 5.0
            instream = port.read(len(refstream))
            if len(instream) != len(refstream):
                raise AssertionError('Stream length differ')
            if instream != refstream:
                raise AssertionError('Stream content differ')
            # the peer expect us to return the same stream, inverted
            outstream = bytes(reversed(instream))
            port.write(outstream)
        finally:
            port.close()

    @classmethod
    def generate_bytes(cls, count=0):
        return ''.join([choice(printable)
                        for x in range(count or cls.COUNT)]).encode()

    @classmethod
    def build_next_url(cls, url):
        iface = int(url[-1])
        iface = (iface + 1) % 3
        return '%s%d' % (url[:-1], iface)


class BaudrateTestCase(FtdiTestCase):
    """Simple test to check clock stretching cannot be overwritten with
       GPIOs.
    """

    BAUDRATES = (300, 600, 1200, 2400, 4800, 9600, 19200, 38400, 57600, 115200,
                 230400, 460800, 576000, 921600, 1000000, 1500000, 1843200,
                 2000000, 2500000, 3000000, 4000000, 6000000, 8000000,
                 12000000)

    def test(self):
        ftdi = Ftdi()
        ftdi.open_from_url(URL)
        for baudrate in self.BAUDRATES:
            actual, _, _ = ftdi._convert_baudrate(baudrate)
            ratio = baudrate/actual
            self.assertTrue(0.97 <= ratio <= 1.03, "Invalid baudrate")


def suite():
    suite_ = TestSuite()
    suite_.addTest(makeSuite(BaudrateTestCase, 'test'))
    suite_.addTest(makeSuite(UartTestCase, 'test'))
    return suite_


def main():
    testmod(modules[__name__])
    FtdiLogger.log.addHandler(logging.StreamHandler(stdout))
    level = environ.get('FTDI_LOGLEVEL', 'info').upper()
    try:
        loglevel = getattr(logging, level)
    except AttributeError as exc:
        raise ValueError('Invalid log level: %s' % level) from exc
    FtdiLogger.set_level(loglevel)
    ut_main(defaultTest='suite')

if __name__ == '__main__':
    if platform == 'darwin':
        # avoid the infamous "The process has forked and you cannot use this
        # CoreFoundation functionality safely. You MUST exec()." error on macOS
        set_start_method('spawn')
    main()



# Copyright (c) 2008-2020, Emmanuel Blot <emmanuel.blot@free.fr>
# Copyright (c) 2008-2016, Neotion
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

# this file has not been updated for a while, so coding style needs some love
#pylint: disable-msg=attribute-defined-outside-init
#pylint: disable-msg=invalid-name
#pylint: disable-msg=missing-class-docstring
#pylint: disable-msg=missing-module-docstring

from io import RawIOBase
from time import sleep, time as now
from serial import SerialBase, SerialException, VERSION as pyserialver
from pyftdi.ftdi import Ftdi
from pyftdi.usbtools import UsbToolsError


class FtdiSerial(SerialBase):
    """Base class for Serial port implementation compatible with pyserial API
       using a USB device.
    """

    BAUDRATES = sorted([9600 * (x+1) for x in range(6)] +
                       list(range(115200, 1000000, 115200)) +
                       list(range(1000000, 13000000, 100000)))

    PYSERIAL_VERSION = tuple([int(x) for x in pyserialver.split('.')])

    def open(self):
        """Open the initialized serial port"""
        if self.port is None:
            raise SerialException("Port must be configured before use.")
        try:
            device = Ftdi.create_from_url(self.port)
        except (UsbToolsError, IOError) as ex:
            raise SerialException('Unable to open USB port %s: %s' %
                                  (self.portstr, str(ex))) from ex
        self.udev = device
        self._set_open_state(True)
        self._reconfigure_port()

    def close(self):
        """Close the open port"""
        self._set_open_state(False)
        if self.udev:
            self.udev.close()
            self.udev = None

    def read(self, size=1):
        """Read size bytes from the serial port. If a timeout is set it may
           return less characters as requested. With no timeout it will block
           until the requested number of bytes is read."""
        data = bytearray()
        start = now()
        while True:
            buf = self.udev.read_data(size)
            data.extend(buf)
            size -= len(buf)
            if size <= 0:
                break
            if self._timeout is not None:
                if buf:
                    break
                ms = now()-start
                if ms > self._timeout:
                    break
            sleep(0.01)
        return bytes(data)

    def write(self, data):
        """Output the given string over the serial port."""
        return self.udev.write_data(data)

    def flush(self):
        """Flush of file like objects. In this case, wait until all data
           is written."""

    def reset_input_buffer(self):
        """Clear input buffer, discarding all that is in the buffer."""
        self.udev.purge_rx_buffer()

    def reset_output_buffer(self):
        """Clear output buffer, aborting the current output and
        discarding all that is in the buffer."""
        self.udev.purge_tx_buffer()

    def send_break(self, duration=0.25):
        """Send break condition."""
        self.udev.set_break(True)
        sleep(duration)
        self.udev.set_break(False)

    def _update_break_state(self):
        """Send break condition. Not supported"""
        self.udev.set_break(self._break_state)

    def _update_rts_state(self):
        """Set terminal status line: Request To Send"""
        self.udev.set_rts(self._rts_state)

    def _update_dtr_state(self):
        """Set terminal status line: Data Terminal Ready"""
        self.udev.set_dtr(self._dtr_state)

    @property
    def ftdi(self) -> Ftdi:
        """Return the Ftdi instance.

           :return: the Ftdi instance
        """
        return self.udev

    @property
    def usb_path(self):
        """Return the physical location as a triplet.
             * bus is the USB bus
             * address is the address on the USB bus
             * interface is the interface number on the FTDI debice

           :return: (bus, address, interface)
           :rtype: tuple(int)
        """
        return self.udev.usb_path

    @property
    def cts(self):
        """Read terminal status line: Clear To Send"""
        return self.udev.get_cts()

    @property
    def dsr(self):
        """Read terminal status line: Data Set Ready"""
        return self.udev.get_dsr()

    @property
    def ri(self):
        """Read terminal status line: Ring Indicator"""
        return self.udev.get_ri()

    @property
    def cd(self):
        """Read terminal status line: Carrier Detect"""
        return self.udev.get_cd()

    @property
    def in_waiting(self):
        """Return the number of characters currently in the input buffer."""
        # not implemented
        return 0

    @property
    def out_waiting(self):
        """Return the number of bytes currently in the output buffer."""
        return 0

    @property
    def fifoSizes(self):
        """Return the (TX, RX) tupple of hardware FIFO sizes"""
        return self.udev.fifo_sizes

    def _reconfigure_port(self):
        try:
            self._baudrate = self.udev.set_baudrate(self._baudrate, True)
            self.udev.set_line_property(self._bytesize,
                                        self._stopbits,
                                        self._parity)
            if self._rtscts:
                self.udev.set_flowctrl('hw')
            elif self._xonxoff:
                self.udev.set_flowctrl('sw')
            else:
                self.udev.set_flowctrl('')
            try:
                self.udev.set_dynamic_latency(12, 200, 50)
            except AttributeError:
                # backend does not support this feature
                pass
        except IOError as exc:
            err = self.udev.get_error_string()
            raise SerialException("%s (%s)" % (str(exc), err)) from exc

    def _set_open_state(self, open_):
        self.is_open = bool(open_)


# assemble Serial class with the platform specific implementation and the base
# for file-like behavior.
class Serial(FtdiSerial, RawIOBase):

    BACKEND = 'pyftdi'

    def __init__(self, *args, **kwargs):
        RawIOBase.__init__(self)
        FtdiSerial.__init__(self, *args, **kwargs)



"""PyUSB virtual FTDI device."""

# Copyright (c) 2020-2021, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

#pylint: disable-msg=missing-docstring
#pylint: disable-msg=unused-argument
#pylint: disable-msg=invalid-name
#pylint: disable-msg=too-many-arguments
#pylint: disable-msg=too-many-locals
#pylint: disable-msg=too-many-branches
#pylint: disable-msg=too-many-statements
#pylint: disable-msg=too-many-instance-attributes
#pylint: disable-msg=too-many-public-methods
#pylint: disable-msg=too-few-public-methods
#pylint: disable-msg=no-self-use

import os
from array import array
from binascii import hexlify
from collections import deque
from enum import IntEnum, unique
from logging import getLogger
from struct import calcsize as scalc, pack as spack, unpack as sunpack
from sys import version_info
from threading import Event, Lock, Thread
from time import sleep, time as now
from typing import List, Mapping, NamedTuple, Optional, Sequence, Tuple
from pyftdi.eeprom import FtdiEeprom   # only for consts, do not use code
from .consts import FTDICONST, USBCONST
from .mpsse import VirtMpsseEngine, VirtMpsseTracer


class Pipe:
    """Wrapper around os.pipe
    """

    PIPE = {'r': 0, 'w': 1}

    def __init__(self):
        self._pipe = None
        self._zombie = False

    def close(self):
        if self._pipe is None:
            return
        os.close(self._pipe[0])
        os.close(self._pipe[1])
        self._pipe = None
        self._zombie = True

    def read(self, count: int) -> bytes:
        return os.read(self.r, count)

    def write(self, buf: bytes) -> None:
        os.write(self.w, buf)

    @property
    def ep_out(self) -> 'Pipe':
        # if pin is output, return RX endpoint (so client can read from)
        return self.r

    @property
    def ep_in(self) -> 'Pipe':
        # if pin in input, return TX endpoint (so client can write to)
        return self.w

    def __getattr__(self, name):
        if self._zombie:
            raise IOError('Closed pipe')
        try:
            pos = self.PIPE[name[0]]
        except KeyError as exc:
            raise AttributeError(f'No such pipe attribute: {name}') from exc
        if not self._pipe:
            # lazy instanciation
            self._pipe = os.pipe()
        return self._pipe[pos]


class Fifo:
    """Communication queue."""

    def __init__(self, size: int = 0):
        self.q = deque()
        self.lock = Lock()
        self.event = Event()
        self.size: int = size


class VirtualFtdiPin:
    """Virtual FTDI pin to enable interconnexion/communication
    """

    class Function(IntEnum):

        TRISTATE = 0
        GPIO = 1
        CLOCK = 2
        STREAM = 3

    def __init__(self, port, position):
        self.log = getLogger(f'pyftdi.vftdi[{port.iface}].{position}')
        self._port = port
        self._position = position
        self._next_pin = None
        self._function = self.Function.TRISTATE
        self._pipe = None

    def connect_to(self, pin: 'VirtualFtdiPin') -> None:
        if self == pin and self._position == pin.position:
            raise ValueError(f'Cannot connect pin {self._port.iface}:'
                             f'{self._position} to itself')
        self._next_pin = pin

    def disconnect(self, pin: 'VirtualFtdiPin') -> None:
        self._next_pin = None

    def set_function(self,
                     function: Optional['VirtualFtdiPin.Function'] = None) \
            -> None:
        self._function = function or self.Function.TRISTATE
        if self._function != self.Function.STREAM:
            if self._pipe:
                self._pipe.close()
                self._pipe = None

    def set(self, source: bool, signal: bool) -> None:
        self.log.debug('set %d <- %s', int(signal), 'ext' if source else 'int')
        if self._function == self.Function.TRISTATE:
            return
        if self._function == self.Function.GPIO:
            if self.direction:
                if self._next_pin:
                    # when propagate to another pin, always an external source
                    self._next_pin.set(True, signal)
            else:
                self._port.set_io(self, signal)
            return
        self.log.error('GPIO update on a featured pin: %s', self._function)

    def read(self, count: int) -> bytes:
        # from external
        if self._function != self.Function.STREAM:
            raise IOError('Pin is not of stream kind')
        if self._pipe is None:
            self._pipe = Pipe()
        return self._pipe.read(count)

    def push_to_pin(self, buf: bytes) -> None:
        # from internal
        if self._function != self.Function.STREAM:
            raise IOError('Pin is not of stream kind')
        if self.is_connected:
            self._next_pin.write(buf)
        if self._pipe is None:
            self._pipe = Pipe()
        self._pipe.write(buf)

    def write(self, buf: bytes) -> None:
        # from external
        if self._function != self.Function.STREAM:
            raise IOError('Pin is not of stream kind')
        if self.is_connected:
            raise IOError(f'Cannot write to connected pin ({self._position})')
        self._port.write_from_pin(self._position, buf)

    @property
    def port(self) -> 'VirtFtdiPort':
        return self._port

    @property
    def connection(self) -> Optional['VirtualFtdiPin']:
        return self._next_pin

    @property
    def is_connected(self) -> bool:
        return bool(self._next_pin)

    @property
    def position(self) -> int:
        return self._position

    @property
    def direction(self) -> bool:
        return bool(self._port.direction & (1 << self._position))

    @property
    def signal(self) -> Optional[bool]:
        if self._function == self.Function.TRISTATE:
            return None
        if self._function == self.Function.GPIO:
            return bool(self._port.gpio & (1 << self._position))
        raise IOError('No signal available from this pin')


class VirtFtdiPort:
    """Virtual FTDI port/interface

       :param iface: the interface number (start from 1)
    """

    POLL_DELAY = 1e-3
    """Delay between processing (generating) a chunk of data."""
    SLEEP_DELAY = 200e-3
    """Timeout when the worker is idle."""

    class Fifos(NamedTuple):
        rx: Fifo  # Host-to-FTDI
        tx: Fifo  # FTDI-to-host


    @unique
    class BitMode(IntEnum):
        """Function mode selection.

           It is used so many times hereafter than relying on FTDICONST is
           not very handy, so redefine it
        """

        RESET = 0x00    # switch off altnerative mode (default to UART)
        BITBANG = 0x01  # classical asynchronous bitbang mode
        MPSSE = 0x02    # MPSSE mode, available on 2232x chips
        SYNCBB = 0x04   # synchronous bitbang mode
        MCU = 0x08      # MCU Host Bus Emulation mode,
        OPTO = 0x10     # Fast Opto-Isolated Serial Interface Mode
        CBUS = 0x20     # Bitbang on CBUS pins of R-type chips
        SYNCFF = 0x40   # Single Channel Synchronous FIFO mode

    FIFO_SIZES = {
        0x0200: (128, 128),    # FT232AM: TX: 128, RX: 128
        0x0400: (128, 384),    # FT232BM: TX: 128, RX: 384
        0x0500: (128, 384),    # FT2232C: TX: 128, RX: 384
        0x0600: (256, 128),    # FT232R:  TX: 256, RX: 128
        0x0700: (4096, 4096),  # FT2232H: TX: 4KiB, RX: 4KiB
        0x0800: (2048, 2048),  # FT4232H: TX: 2KiB, RX: 2KiB
        0x0900: (1024, 1024),  # FT232H:  TX: 1KiB, RX: 1KiB
        0x1000: (512, 512),    # FT-X:    TX: 512, RX: 512
    }
    """FTDI chip internal FIFO sizes.

       TX: to-host, RX: from-host
    """

    WIDTHS = {
        0x0200: 8,
        0x0400: 8,
        0x0500: 12,
        0x0600: 8,
        0x0700: 16,
        0x0800: 8,
        0x0900: 16,
        0x1000: 8}
    """Interterface pin count."""

    UART_PINS = IntEnum('UartPins', 'TXD RXD RTS CTS DTR DSR DCD RI', start=0)
    """Function of pins in UART mode."""

    @unique
    class Command(IntEnum):
        """Command type for command queue (TX thread)."""

        TERMINATE = 0x1    # Terminate thread
        SET_BITMODE = 0x2  # Change the bitmode

    def __init__(self, parent: 'VirtFtdi', iface: int):
        self.log = getLogger(f'pyftdi.vftdi[{iface}]')
        self._parent = parent
        self._iface: int = iface
        self._bitmode = self.BitMode.RESET
        self._hispeed: bool = False
        self._baudrate: int = 0
        self._bb_clock: int = 0
        self._mpsse: Optional[VirtMpsseTracer] = None
        self._direction: int = 0
        self._gpio: int = 0
        self._width = self.WIDTHS[parent.version]
        self._pins: List[VirtualFtdiPin] = []
        self._fifos: VirtFtdiPort.Fifos = VirtFtdiPort.Fifos(
            Fifo(self.FIFO_SIZES[self._parent.version][1]),
            Fifo(self.FIFO_SIZES[self._parent.version][0]))
        self._cbus_dir: int = 0  # logical (commands)
        self._cbus: int = 0  # logical (commands)
        self._cbus_map: Optional[Mapping[int, int]] = None  # logical to phys.
        self._cbusp_gpio: int = 0  # physical (pins)
        self._cbusp_force: int = 0  # physical (pins)
        self._cbusp_active: int = 0  # physical (pins)
        self._resume: bool = True
        self._cmd_q = Fifo()
        self._last_txw_ts = 0
        for pin in range(self._width):
            self._pins.append(VirtualFtdiPin(self, pin))
        self._rx_thread = Thread(
            target=self._rx_worker,
            name=f'Ftdi-{parent.bus}:{parent.address}/{iface}-Rx',
            daemon=True)
        self._tx_thread = Thread(
            target=self._tx_worker,
            name=f'Ftdi-{parent.bus}:{parent.address}/{iface}-Tx',
            daemon=True)
        self._rx_thread.start()
        self._tx_thread.start()

    def close(self, freeze: bool = False) -> None:
        self.log.debug('> close %d', freeze)
        if self._tx_thread:
            with self._cmd_q.lock:
                self._cmd_q.q.append((self.Command.TERMINATE,))
                self._cmd_q.event.set()
        timeout = now()+.5
        while self._cmd_q.q:
            sleep(50e-3)
            if now() >= timeout:
                self.log.warning('aborting before full completion')
                break
        self._resume = False
        if self._rx_thread:
            self._rx_thread.join()
            self._rx_thread = None
        if self._tx_thread:
            self._tx_thread.join()
            self._tx_thread = None

    def terminate(self):
        self.close()

    def __getitem__(self, index: int) -> VirtualFtdiPin:
        if not isinstance(index, int):
            raise TypeError(f"sequence index must be integer, "
                            f"not '{index.__class__.__name__}'")
        try:
            return self._pins[index]
        except IndexError:
            raise IndexError(f'Invalid pin: {index}') from None

    @property
    def iface(self) -> int:
        return self._iface

    @property
    def baudrate(self) -> int:
        return self._baudrate

    @property
    def width(self) -> int:
        return self._width

    @property
    def direction(self) -> int:
        return self._direction

    @property
    def gpio(self) -> int:
        return self._gpio

    @property
    def cbus(self) -> Tuple[int, int]:
        """Emulate CBUS output (from FTDI to peripheral).

           :return: a tuple of logical value on pins, active pins.
                    non-active pins should be considered as High-Z
        """
        return self._cbus_read()

    @cbus.setter
    def cbus(self, cbus: int) -> None:
        """Emulate CBUS input (from peripheral to FTDI)."""
        self._cbus_write(cbus)

    def write(self, data: array, timeout: int) -> int:
        rx_fifo = self._fifos.rx
        with rx_fifo.lock:
            count = len(rx_fifo.q)
            rx_fifo.q.append(data)
            rx_fifo.event.set()
        if not count:
            # the FIFO was empty, so wait for the first request to complete
            # this is a hackish way to ensure a request when the device is
            # not busy handling his FIFOs responds "immediately" to the
            # first request
            timeout = now()+2  # note that verbose traces may trigger a timeout
            while self._resume and rx_fifo.q:
                sleep(self.POLL_DELAY)
                if now() >= timeout:
                    raise RuntimeError('RX queue seems stalled')
        return len(data)

    def read(self, buff: array, timeout: int) -> int:
        tx_fifo = self._fifos.tx
        count = len(buff)
        if count < 2:
            self.log.warning('Never handle buffers that cannot fit status')
            return 0
        if self._bitmode in (self.BitMode.RESET,
                             self.BitMode.BITBANG,
                             self.BitMode.SYNCBB,
                             self.BitMode.MPSSE):
            status = self.modem_status
            buff[0], buff[1] = status[0], status[1]
            pos = 2
            with tx_fifo.lock:
                while tx_fifo.q and pos < count:
                    buff[pos] = tx_fifo.q.popleft()
                    pos += 1
            return pos
        self.log.warning('Read buffer discarded, mode %s',
                         self.BitMode(self._bitmode).name)
        self.log.warning('. bbr (%d)', len(buff))
        return 0

    def set_io(self, pin: VirtualFtdiPin, signal: bool) -> None:
        position = pin.position
        if not 0 <= position < self._width:
            raise ValueError(f'Invalid pin: {position}')
        if signal:
            self._update_gpio(True, self._gpio | (1 << position))
        else:
            self._update_gpio(True, self._gpio & ~(1 << position))

    def update_gpio(self, mpsse: VirtMpsseEngine, source: bool,
                    direction: int, gpio: int) -> None:
        self._direction = direction
        self._update_gpio(source, gpio)

    @property
    def modem_status(self) -> Tuple[int, int]:
        # For some reason, B0 high nibble matches the LPC214x UART:UxMSR
        # B0.0  ?
        # B0.1  ?
        # B0.2  ?
        # B0.3  ?
        # B0.4  Clear to send (CTS)
        # B0.5  Data set ready (DTS)
        # B0.6  Ring indicator (RI)
        # B0.7  Receive line signal / Data carrier detect (RLSD/DCD)

        # For some reason, B1 exactly matches the LPC214x UART:UxLSR
        # B1.0  Data ready (DR)
        # B1.1  Overrun error (OE)
        # B1.2  Parity error (PE)
        # B1.3  Framing error (FE)
        # B1.4  Break interrupt (BI)
        # B1.5  Transmitter holding register (THRE)
        # B1.6  Transmitter empty (TEMT)
        # B1.7  Error in RCVR FIFO
        buf0 = 0x02  # magic constant, no idea for now
        if self._bitmode == self.BitMode.RESET:
            cts = 0x01 if self._gpio & 0x08 else 0
            dsr = 0x02 if self._gpio & 0x20 else 0
            ri = 0x04 if self._gpio & 0x80 else 0
            dcd = 0x08 if self._gpio & 0x40 else 0
            buf0 |= cts | dsr | ri | dcd
        else:
            # another magic constant
            buf0 |= 0x30
        buf1 = 0
        rx_fifo = self._fifos.rx
        with rx_fifo.lock:
            if not rx_fifo.q:
                # TX empty -> flag THRE & TEMT ("TX empty")
                buf1 |= 0x40 | 0x20
        return buf0, buf1

    def control_reset(self, wValue: int, wIndex: int,
                      data: array) -> None:
        reset = FTDICONST.get_name('sio_reset', wValue)
        if reset == 'sio':
            # the thruth is that FTDI does not document what this command
            # exactly does...
            for fifo in self._fifos:
                with fifo.lock:
                    fifo.q.clear()
            self.log.info('> ftdi reset is not fully implemented')
            self._gpio = 0
            self._direction = 0
            self._bitmode = self.BitMode.RESET
            return
        if reset == 'purge_tx':
            fifo = self._fifos.tx
            self.log.info('> ftdi %s: %d requests', reset, len(fifo.q))
            with fifo.lock:
                fifo.q.clear()
            return
        if reset == 'purge_rx':
            fifo = self._fifos.rx
            self.log.info('> ftdi %s: %d requests', reset, len(fifo.q))
            with fifo.lock:
                fifo.q.clear()
            return
        self.log.error('Unknown reset kind: %s', reset)

    def control_set_bitmode(self, wValue: int, wIndex: int,
                            data: array) -> None:
        direction = wValue & 0xff
        bitmode = (wValue >> 8) & 0x7F
        mode = self.BitMode(bitmode).name
        self.log.info('> ftdi bitmode %s: %s', mode, f'{direction:08b}')
        self._bitmode = bitmode
        with self._cmd_q.lock:
            self._cmd_q.q.append((self.Command.SET_BITMODE, self._bitmode))
            self._cmd_q.event.set()
        # be sure to wait for the command to be popped out before resuming
        loop = 10  # is something goes wrong, do not loop forever
        while loop:
            if not self._cmd_q.q:
                # command queue has been fully processed
                break
            if not self._resume:
                # if the worker threads are ending or have ended, abort
                self.log.warning('Premature end of worker')
                return
            loop -= 1
            # kepp some time for the commands to be processed
            sleep(0.05)
        else:
            raise RuntimeError(f'Command {self._cmd_q.q[-1][0].name} '
                               f'not handled')
        if bitmode == self.BitMode.CBUS:
            self._cbus_dir = direction >> 4
            mask = (1 << self._parent.properties.cbuswidth) - 1
            self._cbus_dir &= mask
            # clear output pins
            self._cbus &= ~self._cbus_dir & 0xF
            # update output pins
            output = direction & 0xF & self._cbus_dir
            self._cbus |= output
            self.log.info('> ftdi cbus dir %s, io %s, mask %s',
                          f'{self._cbus_dir:04b}',
                          f'{self._cbus:04b}',
                          f'{mask:04b}')
        elif bitmode == self.BitMode.RESET:
            self._direction = ((1 << VirtFtdiPort.UART_PINS.TXD) |
                               (1 << VirtFtdiPort.UART_PINS.RTS) |
                               (1 << VirtFtdiPort.UART_PINS.DTR))
            self._pins[0].set_function(VirtualFtdiPin.Function.STREAM)
            self._pins[1].set_function(VirtualFtdiPin.Function.STREAM)
            for pin in self._pins[2:]:
                pin.set_function(VirtualFtdiPin.Function.GPIO)
        else:
            self._direction = direction
            for pin in self._pins:
                pin.set_function(VirtualFtdiPin.Function.GPIO)
        if bitmode == self.BitMode.MPSSE:
            for pin in self._pins:
                pin.set_function(VirtualFtdiPin.Function.GPIO)
            if not self._mpsse:
                self._mpsse = VirtMpsseTracer(self, self._parent.version)

    def control_set_latency_timer(self, wValue: int, wIndex: int,
                                  data: array) -> None:
        self.log.debug('> ftdi latency timer: %d', wValue)

    def control_set_event_char(self, wValue: int, wIndex: int,
                               data: array) -> None:
        char = wValue & 0xFF
        enable = bool(wValue >> 8)
        self.log.info('> ftdi %sable event char: 0x%02x',
                      'en' if enable else 'dis', char)

    def control_set_error_char(self, wValue: int, wIndex: int,
                               data: array) -> None:
        char = wValue & 0xFF
        enable = bool(wValue >> 8)
        self.log.info('> ftdi %sable error char: 0x%02x',
                      'en' if enable else 'dis', char)

    def control_read_pins(self, wValue: int, wIndex: int,
                          data: array) -> bytes:
        mode = FTDICONST.get_name('bitmode', self._bitmode)
        self.log.info('> ftdi read_pins %s', mode)
        if mode == 'cbus':
            cbus = self._cbus & ~self._cbus_dir & 0xF
            self.log.info('< cbus 0x%01x: %s', cbus, f'{cbus:04b}')
            return bytes([cbus])
        low_gpio = self._gpio & 0xFF
        self.log.info('< gpio 0x%02x: %s', low_gpio, f'{low_gpio:08b}')
        return bytes([low_gpio])

    def control_set_baudrate(self, wValue: int, wIndex: int,
                             data: array) -> None:
        FRAC_INV_DIV = (0, 4, 2, 1, 3, 5, 6, 7)
        BAUDRATE_REF_BASE = 3.0E6  # 3 MHz
        BAUDRATE_REF_HIGH = 12.0E6  # 12 MHz
        if self._parent.is_hispeed_device or self._parent.is_x_series:
            wIndex >>= 8
        divisor = wValue | (wIndex << 16)
        div = divisor & 0x3FFF
        hispeed = bool(divisor & 0x20000)
        if not self._parent.is_hispeed_device and hispeed:
            raise ValueError('Invalid hispeed mode with non-H series')
        subdiv_code = (divisor >> 14) & 0x7
        refclock = BAUDRATE_REF_HIGH if hispeed else BAUDRATE_REF_BASE
        if div < 2 and subdiv_code:
            raise ValueError('Invalid sub-divisor with special div')
        if div == 0:
            baudrate = 3.0e6
        elif div == 1:
            baudrate = 2.0e6
        else:
            subdiv = FRAC_INV_DIV[subdiv_code]
            baudrate = round((refclock * 8) / (div * 8 + subdiv))
        self._hispeed = hispeed
        self._baudrate = int(baudrate)
        multiplier = FTDICONST.get_value('BITBANG_BAUDRATE_RATIO',
                                         'HIGH' if hispeed else 'BASE')
        self._bb_clock = self._baudrate * multiplier
        self.log.info('> ftdi set_baudrate %d bps, HS: %s, bb %d Hz',
                      baudrate, hispeed, self._bb_clock)

    def control_set_data(self, wValue: int, wIndex: int,
                         data: array) -> None:
        self.log.debug('> ftdi set_data (NOT IMPLEMENTED)')

    def control_set_flow_ctrl(self, wValue: int, wIndex: int,
                              data: array) -> None:
        self.log.debug('> ftdi set_flow_ctrl (NOT IMPLEMENTED)')

    def control_poll_modem_status(self, wValue: int, wIndex: int,
                                  data: array) -> None:
        status = self.modem_status
        self.log.info('> ftdi poll_modem_status %02x%02x',
                      status[0], status[1])
        return status

    def write_from_pin(self, pin: int, buf: bytes) -> None:
        tx_fifo = self._fifos.tx
        with tx_fifo.lock:
            free_count = tx_fifo.size - len(tx_fifo.q)
            if free_count > 0:
                tx_fifo.q.extend(buf[:free_count])
        if free_count < len(buf):
            self.log.warning('FIFO full, truncated buffer from %d', pin)

    def write_from_mpsse(self, mpsse: VirtMpsseEngine, buf: bytes) -> None:
        tx_fifo = self._fifos.tx
        with tx_fifo.lock:
            free_count = tx_fifo.size - len(tx_fifo.q)
            if free_count > 0:
                tx_fifo.q.extend(buf[:free_count])
        if free_count < len(buf):
            self.log.warning('FIFO full (%d bytes), truncated buffer',
                             len(tx_fifo.q))
        self._mpsse.receive(self._iface, buf)

    def _update_gpio(self, source: bool, gpio: int) -> None:
        changed = self._gpio ^ gpio
        self._gpio = gpio
        if source:
            # call comes from an external source, no need to propagate more
            return
        for pos, pin in enumerate(self._pins):
            bit = 1 << pos
            if not bit & self._direction:
                # do not update input pins
                continue
            if bit & changed:
                pin.set(False, bool(gpio & (1 << pos)))

    def _decode_cbus_x1000(self) -> None:
        cbus_gpio = 0
        cbus_force = 0
        cbus_active = 0
        for bix in range(4):
            value = self._parent.eeprom[0x1A + bix]
            bit = 1 << bix
            if FtdiEeprom.CBUSX(value).name == 'GPIO':
                cbus_gpio |= bit
                cbus_active |= bit
            elif FtdiEeprom.CBUSX(value).name == 'DRIVE0':
                cbus_force &= ~bit  # useless, for code symmetry
                cbus_active |= bit
            elif FtdiEeprom.CBUSX(value).name == 'DRIVE1':
                cbus_force |= bit
                cbus_active |= bit
        mask = (1 << self._parent.properties.cbuswidth) - 1
        self._cbusp_gpio = cbus_gpio & mask
        self._cbusp_force = cbus_force & mask
        self._cbusp_active = cbus_active & mask
        self.log.debug('x1000 config gpio %s, force %s, active %s',
                       f'{self._cbusp_gpio:04b}',
                       f'{self._cbusp_force:04b}',
                       f'{self._cbusp_active:04b}')

    def _decode_cbus_x0900(self) -> None:
        cbus_gpio = 0
        cbus_force = 0
        cbus_active = 0
        for bix in range(5):
            value = self._parent.eeprom[0x18 + bix]
            low, high = value & 0x0F, value >> 4
            bit = 1 << 2*bix
            if FtdiEeprom.CBUSH(low).name == 'GPIO':
                cbus_gpio |= bit
                cbus_active |= bit
            elif FtdiEeprom.CBUSH(low).name == 'DRIVE0':
                cbus_force &= ~bit  # useless, for code symmetry
                cbus_active |= bit
            elif FtdiEeprom.CBUSH(low).name == 'DRIVE1':
                cbus_force |= bit
                cbus_active |= bit
            bit <<= 1
            if FtdiEeprom.CBUSH(high).name == 'GPIO':
                cbus_gpio |= bit
                cbus_active |= bit
            elif FtdiEeprom.CBUSH(high).name == 'DRIVE0':
                cbus_force &= ~bit  # useless, for code symmetry
                cbus_active |= bit
            elif FtdiEeprom.CBUSH(high).name == 'DRIVE1':
                cbus_force |= bit
                cbus_active |= bit
        mask = (1 << self._parent.properties.cbuswidth) - 1
        self._cbusp_gpio = cbus_gpio & mask
        self._cbusp_force = cbus_force & mask
        self._cbusp_active = cbus_active & mask
        self._cbus_map = {0: 5, 1: 6, 2: 8, 3: 9}
        self.log.debug('x0900 config gpio %s, force %s, active %s',
                       f'{self._cbusp_gpio:04b}',
                       f'{self._cbusp_force:04b}',
                       f'{self._cbusp_active:04b}')

    def _decode_cbus_x0600(self) -> None:
        cbus_gpio = 0
        cbus_active = 0
        bix = 0
        while True:
            value = self._parent.eeprom[0x14 + bix]
            low, high = value & 0x0F, value >> 4
            bit = 1 << (2*bix)
            if FtdiEeprom.CBUS(low).name == 'GPIO':
                cbus_gpio |= bit
                cbus_active |= bit
            if bix == 2:
                break
            bit <<= 1
            if FtdiEeprom.CBUS(high).name == 'GPIO':
                cbus_gpio |= bit
                cbus_active |= bit
            bix += 1
        mask = (1 << self._parent.properties.cbuswidth) - 1
        self._cbusp_gpio = cbus_gpio & mask
        self._cbusp_force = 0
        self._cbusp_active = cbus_active & mask
        self.log.debug('x0600 config gpio %s, force %s, active %s',
                       f'{self._cbusp_gpio:04b}',
                       f'{self._cbusp_force:04b}',
                       f'{self._cbusp_active:04b}')

    def _cbus_write(self, cbus: int) -> None:
        # from peripheral to FTDI
        # mask out CBUS pins which are not configured as GPIOs
        cbus &= self._cbusp_active
        cbus &= self._cbusp_gpio
        self.log.debug('> cbus_write active: %s gpio: %s, force: %s, cbus: %s',
                       f'{self._cbusp_active:04b}',
                       f'{self._cbusp_gpio:04b}',
                       f'{self._cbusp_force:04b}',
                       f'{cbus:04b}')
        if self._cbus_map:
            self.log.info('cbus_write map')
            # convert physical gpio into logical gpio
            lgpio = 0
            for log, phy in self._cbus_map:
                if cbus & (1 << phy):
                    lgpio |= 1 << log
            cbus = lgpio
        # only consider logical input
        cbus &= ~self._cbus_dir
        # combine existing output with new input
        self._cbus &= ~self._cbus_dir
        self._cbus |= cbus & ~self._cbus_dir

    def _cbus_read(self) -> Tuple[int, int]:
        # from FTDI to peripheral
        cbus = self._cbus
        self.log.debug('> cbus_read active %s, gpio %s, force %s, cbus %s',
                       f'{self._cbusp_active:04b}',
                       f'{self._cbusp_gpio:04b}',
                       f'{self._cbusp_force:04b}',
                       f'{cbus:04b}')
        if self._cbus_map:
            self.log.info('cbus_read map')
            # convert logical gpio into physical gpio
            pgpio = 0
            for log, phy in self._cbus_map:
                if cbus & (1 << log):
                    pgpio |= 1 << phy
            cbus = pgpio
        # mask out CBUS pins which are not configured as GPIOs
        cbus &= self._cbusp_gpio
        # apply DRIVE1 to gpio
        cbus |= self._cbusp_force
        self.log.info('< cbus_read cbus %s, active %s',
                      f'{cbus:04b}', f'{self._cbusp_active:04b}')
        return cbus, self._cbusp_active

    def _rx_worker(self):
        """Background handling of data received from host."""
        try:
            while self._resume:
                rx_fifo = self._fifos.rx
                rx_fifo.lock.acquire()
                if not rx_fifo.q:
                    rx_fifo.lock.release()
                    if not rx_fifo.event.wait(self.POLL_DELAY):
                        continue
                    rx_fifo.lock.acquire()
                    rx_fifo.event.clear()
                    if not rx_fifo.q:
                        self.log.error('wake up w/o RX data')
                        rx_fifo.lock.release()
                        continue
                rx_fifo.event.clear()
                data = rx_fifo.q[0]
                rx_fifo.lock.release()
                if self._bitmode == self.BitMode.MPSSE:
                    self._mpsse.send(self._iface, data)
                elif self._bitmode == self.BitMode.RESET:
                    self[self.UART_PINS.TXD].push_to_pin(data)
                elif self._bitmode == self.BitMode.BITBANG:
                    for byte in data:
                        # only 8 LSBs are addressable through this command
                        gpi = self._gpio & ~self._direction & 0xFF
                        gpo = byte & self._direction & 0xFF
                        msb = self._gpio & ~0xFF
                        gpio = gpi | gpo | msb
                        self._update_gpio(False, gpio)
                        self.log.debug('. bbw %02x: %s',
                                       self._gpio, f'{self._gpio:08b}')
                elif self._bitmode == self.BitMode.SYNCBB:
                    tx_fifo = self._fifos.tx
                    lost = 0
                    for byte in data:
                        with tx_fifo.lock:
                            free_count = tx_fifo.size - len(tx_fifo.q)
                            if free_count > 0:
                                tx_fifo.q.append(self._gpio & 0xFF)
                            else:
                                lost += 1
                        # only 8 LSBs are addressable through this command
                        gpi = self._gpio & ~self._direction & 0xFF
                        gpo = byte & self._direction & 0xFF
                        msb = self._gpio & ~0xFF
                        gpio = gpi | gpo | msb
                        self._update_gpio(False, gpio)
                        self.log.debug('. bbw %02x: %s',
                                       self._gpio, f'{self._gpio:08b}')
                    if lost:
                        self.log.debug('%d samples lost, TX full', lost)
                else:
                    try:
                        mode = self.BitMode(self._bitmode).name
                    except ValueError:
                        mode = 'unknown'
                    self.log.warning('Write buffer discarded, mode %s', mode)
                    self.log.warning('. (%d) %s',
                                     len(data), hexlify(data).decode())
                with rx_fifo.lock:
                    rx_fifo.q.popleft()
            self.log.debug('End of worker %s', self._rx_thread.name)
        except Exception as exc:
            self.log.error('Dead of worker %s: %s', self._rx_thread.name, exc)
            raise
        finally:
            # ensure other threads to not run forever if this one is on error
            if self._resume:
                self.log.info('RX worker is triggering death')
            self._resume = False

    def _tx_worker(self):
        """Background handling of data sent to host."""
        bitmode = self.BitMode.RESET
        _wait_delay = self.SLEEP_DELAY
        terminated = False
        try:
            while self._resume:
                # in order not to overload CPU with real-time computation,
                # time is sliced into POLL_DELAY period; every POLL_DELAY
                # period, this thead generates as many data as the real HW
                # would have generated during this period (~ epsilon due to
                # rounding)
                if not self._cmd_q.q:
                    if not self._cmd_q.event.wait(_wait_delay):
                        self._tx_worker_generate(bitmode)
                        continue
                with self._cmd_q.lock:
                    if not self._cmd_q.q:
                        self.log.error('wake up w/o CMD')
                        continue
                    self._cmd_q.event.clear()
                    command = self._cmd_q.q.popleft()
                self.log.info('Command %s', self.Command(command[0]).name)
                if command[0] == self.Command.TERMINATE:
                    terminated = True
                    break
                if command[0] == self.Command.SET_BITMODE:
                    bitmode = command[1]
                    self._last_txw_ts = now()
                    if bitmode in (self.BitMode.BITBANG,
                                   self.BitMode.SYNCBB):
                        _wait_delay = self.POLL_DELAY
                    else:
                        _wait_delay = self.SLEEP_DELAY
                    continue
                else:
                    self.log.error('Unimplemented support for command %d',
                                   command)
                    continue
            self.log.debug('End of worker %s', self._tx_thread.name)
        except Exception as exc:
            self.log.error('Dead of worker %s: %s', self._tx_thread.name, exc)
            raise
        finally:
            # ensure other threads to not run forever if this one is on error
            if not terminated and self._resume:
                self.log.info('TX worker is triggering death')
            self._resume = False

    def _tx_worker_generate(self, bitmode) -> None:
        tx_fifo = self._fifos.tx
        if bitmode == self.BitMode.BITBANG:
            ts = now()
            # how much time has elapsed since last background action
            elapsed = ts-self._last_txw_ts
            self._last_txw_ts = ts
            # how many bytes should have been captured since last
            # action
            byte_count = round(self._baudrate*elapsed)
            # fill the TX FIFO with as many bytes, stop if full
            if byte_count:
                with tx_fifo.lock:
                    free_count = tx_fifo.size - len(tx_fifo.q)
                    push_count = min(free_count, byte_count)
                    tx_fifo.q.extend([self._gpio] * push_count)
                self.log.debug('in %.3fms -> %d',
                               elapsed*1000, push_count)


class VirtFtdi:
    """Virtual FTDI device.

       :param version: FTDI version (device kind)
       :param eeprom_size: size of external EEPROM size, if any
    """

    class Properties(NamedTuple):
        """Device properties."""
        ifcount: int  # count of interface
        ifwidth: int  # pin width of an interface
        cbuswidth: int  # pin width of the control bus

    EXT_EEPROMS: Mapping[str, int] = {
        '93c46': 128,  # 1024 bits
        '93c56': 256,  # 2048 bits
        '93c66': 256,  # 2048 bits  (93C66 seen as 93C56)
    }
    """External EEPROMs."""

    INT_EEPROMS: Mapping[int, int] = {
        0x0600: 0x80,  # FT232R: 128 bytes, 1024 bits
        0x1000: 0x400  # FT23*X: 1KiB
    }
    """Internal EEPROMs."""

    PROPERTIES: Mapping[int, Properties] = {
        0x0200: Properties(1, 8, 0),   # FT232AM
        0x0400: Properties(1, 8, 0),   # FT232BM
        0x0500: Properties(1, 8, 0),   # FT2232D
        0x0600: Properties(1, 8, 5),   # FT232R
        0x0700: Properties(2, 16, 0),  # FT2232H
        0x0800: Properties(4, 8, 0),   # FT4232H
        0x0900: Properties(1, 8, 10),  # FT232H
        0x1000: Properties(1, 8, 4),   # FT231X
    }
    """Width of port/bus (regular, cbus)."""

    EEPROM_REQ_BASE = FTDICONST.get_value('SIO_REQ', 'EEPROM')
    """Base value for EEPROM request."""

    def __init__(self, version: int, bus: int, address: int,
                 eeprom: Optional[dict] = None):
        self.log = getLogger('pyftdi.vftdi')
        self._version = version
        self._bus: int = bus
        self._address: int = address
        self._eeprom: bytearray = self._build_eeprom(version, eeprom)
        self._ports: List[VirtFtdiPort] = []
        for iface in range(self.PROPERTIES[self._version].ifcount):
            self._ports.append(VirtFtdiPort(self, iface+1))

    def close(self, freeze: bool = False) -> None:
        for port in self._ports:
            port.close(freeze)

    def terminate(self):
        self.close()

    @property
    def version(self) -> int:
        return self._version

    @property
    def bus(self) -> int:
        return self._bus

    @property
    def address(self) -> int:
        return self._address

    @property
    def properties(self) -> 'VirtFtdi.Properties':
        return self.PROPERTIES[self._version]

    @property
    def is_mpsse_device(self) -> bool:
        """Tell whether the device has an MPSSE engine

           :return: True if the FTDI device supports MPSSE
        """
        return self._version in (0x0500, 0x0700, 0x0800, 0x0900)

    @property
    def is_hispeed_device(self) -> bool:
        """Tell whether the device is a high speed device

           :return: True if the FTDI device is HS
        """
        return self._version in (0x0700, 0x0800, 0x0900)

    @property
    def is_x_series(self) -> bool:
        """Tell whether the device is a FT-X device

           :return: True for FT-X device
        """
        return self._version == 0x1000

    def apply_eeprom_config(self, devdesc: dict,
                            cfgdescs: Sequence[dict]) -> None:
        self._load_eeprom(devdesc, cfgdescs)

    def control(self, dev_handle: 'VirtDeviceHandle', bmRequestType: int,
                bRequest: int, wValue: int, wIndex: int, data: array,
                timeout: int) -> int:
        req_ctrl = USBCONST.dec_req_ctrl(bmRequestType)
        req_type = USBCONST.dec_req_type(bmRequestType)
        req_rcpt = USBCONST.dec_req_rcpt(bmRequestType)
        req_desc = ':'.join([req_ctrl, req_type, req_rcpt])
        req_name = FTDICONST.dec_req_name(bRequest)
        dstr = (hexlify(data).decode() if USBCONST.is_req_out(bmRequestType)
                else f'({len(data)})')
        self.log.debug('> control ftdi hdl %d, %s, %s, '
                       'val 0x%04x, idx 0x%04x, data %s, to %d',
                       dev_handle.handle, req_desc, req_name,
                       wValue, wIndex, dstr, timeout)
        size = 0
        try:
            if bRequest >= self.EEPROM_REQ_BASE:
                obj = self
            else:
                obj = self._ports[0xFF & (wIndex-1)] if wIndex else self
        except IndexError:
            raise ValueError(f'Invalid iface: 0x{wIndex:04x} '
                             f'for {req_name}') from None
        try:
            pre = '_' if obj == self else ''
            handler = getattr(obj, f'{pre}control_{req_name}')
        except AttributeError:
            self.log.warning('Unknown request %02x:%04x: %s for %s)',
                             bRequest, wIndex, req_name,
                             obj.__class__.__name__)
            return size
        buf = handler(wValue, wIndex, data) or b''
        size = len(buf)
        data[:size] = array('B', buf)
        self.log.debug('< (%d) %s', size, hexlify(data[:size]).decode())
        return size

    def write(self, dev_handle: 'VirtDeviceHandle', ep: int, intf: int,
              data: array, timeout: int) -> int:
        return self._ports[intf].write(data, timeout)

    def read(self, dev_handle: 'VirtDeviceHandle', ep: int, intf: int,
             buff: array, timeout: int) -> int:
        return self._ports[intf].read(buff, timeout)

    def get_port(self, iface: int):
        # iface: 1..n
        return self._ports[iface-1]

    @property
    def eeprom(self) -> bytes:
        return bytes(self._eeprom)

    @eeprom.setter
    def eeprom(self, value: bytes):
        if len(value) != len(self._eeprom):
            raise ValueError('EEPROM size mismatch')
        self._eeprom = bytearray(value)

    @classmethod
    def _build_eeprom(cls, version, eeprom: Optional[dict]) -> bytearray:
        size = 0
        data = b''
        if eeprom:
            model = eeprom.get('model', None)
            if model:
                if version in cls.INT_EEPROMS:
                    raise ValueError('No external EEPROM supported on this '
                                     'device')
                try:
                    size = cls.EXT_EEPROMS[model.lower()]
                except KeyError as exc:
                    raise ValueError('Unsupported EEPROM model: {model}') \
                            from exc
            data = eeprom.get('data', b'')
        if version in cls.INT_EEPROMS:
            int_size = cls.INT_EEPROMS[version]
            # FT232R, FT230x, FT231x, FT234x
            if size:
                if size != int_size:
                    raise ValueError('Internal EEPROM size cannot be changed')
            else:
                size = int_size
        else:
            if size and size not in cls.EXT_EEPROMS.values():
                raise ValueError(f'Invalid EEPROM size: {size}')
        if data and len(data) > size:
            raise ValueError('Data cannot fit into EEPROM')
        buf = bytearray(size)
        buf[:len(data)] = data
        return buf

    def _checksum_eeprom(self, data: bytearray) -> int:
        length = len(data)
        if length & 0x1:
            raise ValueError('Length not even')
        # NOTE: checksum is computed using 16-bit values in little endian
        # ordering
        checksum = 0XAAAA
        mtp = self._version == 0x1000  # FT230X
        for idx in range(0, length, 2):
            if mtp and 0x24 <= idx < 0x80:
                # special MTP user section which is not considered for the CRC
                continue
            val = ((data[idx+1] << 8) + data[idx]) & 0xffff
            checksum = val ^ checksum
            checksum = ((checksum << 1) & 0xffff) | ((checksum >> 15) & 0xffff)
        return checksum

    def _load_eeprom(self, devdesc: dict, cfgdescs: Sequence[dict]) -> None:
        whole = self._version != 0x1000  # FT230X
        buf = self._eeprom if whole else self._eeprom[:0x100]
        chksum = self._checksum_eeprom(buf)
        if chksum:
            self.log.warning('Invalid EEPROM checksum, ignoring content')
            return
        # only apply a subset of what the EEPROM can configure for now
        devdesc['idVendor'] = sunpack('<H', self._eeprom[2:4])[0]
        devdesc['idProduct'] = sunpack('<H', self._eeprom[4:6])[0]
        devdesc['iManufacturer'] = self._decode_eeprom_string(0x0e)
        devdesc['iProduct'] = self._decode_eeprom_string(0x10)
        devdesc['iSerialNumber'] = self._decode_eeprom_string(0x12)
        for desc in cfgdescs:
            if desc.bConfigurationValue == 0:
                # only update first configuration
                desc['bMaxPower'] = self._eeprom[0x09]
                desc['bmAttributes'] = 0x80 | (self._eeprom[0x08] & 0x0F)
        cbus_dec = f'_decode_cbus_x{self._version:04x}'
        try:
            cbus_func = getattr(self._ports[0], cbus_dec)
        except AttributeError:
            self.log.debug('No CBUS support: %s', cbus_dec)
            return
        cbus_func()

    def _decode_eeprom_string(self, offset):
        str_offset, str_size = sunpack('<BB', self._eeprom[offset:offset+2])
        if str_size:
            str_size -= scalc('<H')
            str_offset += scalc('<H')
            manufacturer = self._eeprom[str_offset:str_offset+str_size]
            return manufacturer.decode('utf16', errors='ignore')
        return ''

    def _control_read_eeprom(self, wValue: int, wIndex: int,
                             data: array) -> Optional[bytes]:
        self.log.debug('> ftdi read_eeprom @ 0x%04x', wIndex*2)
        if not self._eeprom:
            self.log.warning('Missing EEPROM')
            return None
        if len(data) != 2:
            self.log.warning('Unexpected read size: %d', len(data))
            # resume anyway
        address = abs(wIndex * 2)
        if address + 1 > len(self._eeprom):
            # out of bound
            self.log.warning('Invalid EEPROM address: 0x%04x', wValue)
            return None
        word = bytes(self._eeprom[address: address+2])
        return word

    def _control_write_eeprom(self, wValue: int, wIndex: int,
                              data: array) -> None:
        self.log.info('> ftdi write_eeprom @ 0x%04x', wIndex*2)
        if not self._eeprom:
            self.log.warning('Missing EEPROM')
            return
        address = abs(wIndex * 2)
        if address + 1 > len(self._eeprom):
            # out of bound
            self.log.warning('Invalid EEPROM address: 0x%04x', wValue)
            return
        if self._version == 0x1000:
            if 0x80 <= address < 0xA0:
                # those address are R/O on FT230x
                self.log.warning('Protected EEPROM address: 0x%04x', wValue)
                return
        self._eeprom[address: address+2] = spack('<H', wValue)

    def _control_set_baudrate(self, wValue: int, wIndex: int,
                              data: array) -> None:
        # set_baudrate is very specific and hack-ish, as the HW is.
        # the way the interface is encoded is simply ugly (likely a legacy
        # from single-port device, then FTDI needed to support multi-port
        # ones and found a hack-ish way to implement it w/o breaking the
        # existing implementation)
        if self.is_mpsse_device:
            iface = wIndex & 0xFF
            wIndex >>= 16
        else:
            iface = 1
        return self._ports[iface-1].control_set_baudrate(wValue, wIndex, data)
